#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Namespace.

A namespace is nothing more than a dictionary with some benefits:

* items use `item.name` as dictionary key. This is checked.
* items **cannot** be overwritten
* items **cannot** be deleted
* the namespace can be locked for modifications via `lock`.
* iteration over the namespaces yields the items and not their keys.

??? Example "Namespace Examples"
    Examples.

        >>> from collections import namedtuple
        >>> NamedObject = namedtuple('NamedObject', 'name foo bar')
        >>> namespace = Namespace([NamedObject('a', 1, 2)])
        >>> namespace.add(NamedObject('b', 3, 4))
        >>> namespace['c'] = NamedObject('c', 5, 6)
        >>> namespace.is_locked
        False
        >>> for item in namespace:
        ...     item
        NamedObject(name='a', foo=1, bar=2)
        NamedObject(name='b', foo=3, bar=4)
        NamedObject(name='c', foo=5, bar=6)
        >>> for item in namespace.keys():
        ...     item
        'a'
        'b'
        'c'
        >>> for item in namespace.values():
        ...     item
        NamedObject(name='a', foo=1, bar=2)
        NamedObject(name='b', foo=3, bar=4)
        NamedObject(name='c', foo=5, bar=6)
        >>> namespace['b']
        NamedObject(name='b', foo=3, bar=4)
        >>> namespace['d']
        Traceback (most recent call last):
        ...
        KeyError: 'd'
        >>> namespace.get_dym('d')
        Traceback (most recent call last):
        ...
        ValueError: 'd'. Known are 'a', 'b' and 'c'.

    Locking:

        >>> namespace.lock()
        >>> namespace.is_locked
        True
        >>> namespace['d'] = NamedObject('d', 7, 8)
        Traceback (most recent call last):
        ...
        ucdp.exceptions.LockError: Namespace is already locked. Cannot add items anymore.

        >>> len(namespace)
        3
"""

from collections.abc import Iterable
from typing import Any

from .exceptions import DuplicateError, LockError
from .nameutil import didyoumean
from .object import NamedObject


class Namespace(dict):
    """
    Namespace.
    """

    def __init__(self, items: Iterable[NamedObject] | None = None):
        super().__init__()
        self.__is_locked = False
        if items:
            for item in items:
                self[item.name] = item

    @property
    def is_locked(self) -> bool:
        """Locked."""
        return self.__is_locked

    def lock(self) -> None:
        """Lock."""
        if self.__is_locked:
            raise LockError("Namespace is already locked. Cannot lock again.")
        self.__is_locked = True

    def add(self, item: NamedObject, exist_ok: bool = False):
        """Add."""
        try:
            self[item.name] = item
        except DuplicateError as exc:
            if not exist_ok:
                raise exc

    def get_dym(self, name: str) -> NamedObject:
        """Get NamedObject."""
        try:
            item = self[name]
        except KeyError as exc:
            dym = didyoumean(name, self.keys(), known=True)
            raise ValueError(f"{exc!s}.{dym}") from None
        return item

    def __setitem__(self, name, item):
        self._set_items(((name, item),))

    def _set_items(self, items: Iterable[tuple[str, NamedObject]]):
        if self.__is_locked:
            raise LockError("Namespace is already locked. Cannot add items anymore.")
        for name, item in items:
            if item.name != name:
                raise ValueError(f"{item} with must be stored at name '{item.name}' not at '{name}'")
            if item.name in self.keys():
                if item is self[item.name]:
                    raise DuplicateError(f"{self[item.name]!r} already exists")
                raise DuplicateError(f"Name '{item.name}' already taken by {self[item.name]!r}")
            super().__setitem__(name, item)

    def __delitem__(self, key):
        raise TypeError(f"It is forbidden to remove {key!r}.")

    def __iter__(self):
        yield from self.values()

    def __repr__(self):
        return f"{self.__class__.__qualname__}({list(self)})"

    def pop(self, *args) -> Any:
        """
        Pop value.

        If key is in the dictionary, remove it and return its value, else return default.
        If default is not given and key is not in the dictionary, a KeyError is raised.

        This operation is forbidden.
        """
        raise TypeError("It is forbidden to remove any item.")

    def popitem(self) -> tuple[Any, Any]:
        """
        Remove and return a (key, value) pair from the dictionary.

        This operation is forbidden.
        """
        raise TypeError("It is forbidden to remove any item.")

    def set_default(self, key, value=None) -> Any:
        """
        Set Default.

        If key is in the dictionary, return its value.
        If not, insert key with a value of default and return default. ``default`` defaults to None.
        """
        if self.__is_locked:
            raise LockError("Namespace is already locked. Cannot add items anymore.")
        if key in self.keys():
            return self[key]
        self[key] = value
        return value

    def update(self, other: dict) -> None:  # type: ignore[override]
        """Update the dictionary with the key/value pairs from other."""
        self._set_items(other.items())

    def __ior__(self, other):
        if not isinstance(other, dict):
            return NotImplemented
        self._set_items(other.items())
        return self

    def __or__(self, other) -> "Namespace":
        if not isinstance(other, dict):
            return NotImplemented
        if self.__is_locked:
            raise LockError("Namespace is already locked. Cannot add items anymore.")
        items = dict(self)
        items.update(other)
        namespace = self.__class__()
        namespace._set_items(items.items())
        return namespace



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Testbench Module.
"""

from typing import Any, ClassVar

from caseconverter import snakecase

from ._modbuilder import build
from .modbase import BaseMod, ModClss
from .modtb import ATbMod
from .modutil import get_modname


class AGenericTbMod(ATbMod):
    """
    A Generic Testbench which adapts to dut.

    Create testbench for `dut`.

    The [TopModRef][ucdp.modtopref.TopModRef] and [load()][ucdp.loader.load] function
    allow to wrap any design module with a testbench.

    ??? Example "Module Testbench Examples"
        Example:

            >>> import ucdp as u
            >>> class MyMod(u.AMod):
            ...
            ...     def _build(self):
            ...         self.add_port(u.UintType(4), "data_i")
            ...         self.add_port(u.UintType(4), "data_o")
            ...
            >>> class OtherMod(u.AMod):
            ...
            ...     def _build(self):
            ...         self.add_port(u.UintType(4), "data_i")
            ...         self.add_port(u.UintType(4), "data_o")
            ...
            >>> class GenTbMod(u.AGenericTbMod):
            ...
            ...     def _build(self):
            ...         # Build testbench for self.dut here
            ...         pass
            ...
            ...     @staticmethod
            ...     def build_dut():
            ...         return MyMod()

            >>> tb = GenTbMod()
            >>> tb
            <ucdp.modgenerictb.GenTbMod(inst='gen_tb_my', libname='ucdp', modname='gen_tb_my', dut=...)>)>
            >>> tb.dut
            <ucdp.modgenerictb.MyMod(inst='my', libname='ucdp', modname='my')>
            >>> tb = GenTbMod(OtherMod())
            >>> tb
            <ucdp.modgenerictb.GenTbMod(inst='gen_tb_other', libname='ucdp', modname='...', dut=...)>)>
            >>> tb.dut
            <ucdp.modgenerictb.OtherMod(inst='other', libname='ucdp', modname='other')>

    [TopModRef][ucdp.modtopref.TopModRef] and [load()][ucdp.loader.load] handle that gracefully and
    allow pairing of testbench and dut on command line and in configuration files.
    """

    dut_modclss: ClassVar[ModClss] = set()
    """Testbench is limited to these kind of modules."""
    dut: BaseMod

    def __init__(self, dut: BaseMod | None = None, name: str | None = None, **kwargs):
        cls = self.__class__
        if dut is None:
            dut = cls.build_dut()
        if not name:
            basename = snakecase(cls.__name__.removesuffix("Mod"))
            name = f"{basename}_{dut.modname}"
        dut_modclss: tuple[type[BaseMod], ...] = tuple(cls.dut_modclss)
        if dut_modclss:
            if not isinstance(dut, dut_modclss):
                raise TypeError(f"{cls} can only test {dut_modclss} modules, but not {dut.__class__} module")
        super().__init__(parent=None, name=name, dut=dut, **kwargs)  # type: ignore[call-arg]

    @property
    def modname(self) -> str:
        """Module Name."""
        modbasename = get_modname(self.__class__)
        return f"{modbasename}_{self.dut.modname}"

    @classmethod
    def build_tb(cls, dut, **kwargs) -> "ATbMod":
        """Build Testbench."""
        return cls(dut, **kwargs)

    @classmethod
    def build_dut(cls, **kwargs) -> BaseMod:
        """Build DUT."""
        raise NotImplementedError

    def _build(self) -> None:
        """Build."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        self.add_inst(self.dut)
        self._build()
        build(self)

    def __str__(self):
        modref = self.get_modref()
        return f"<{modref}(inst={self.inst!r}, libname={self.libname!r}, modname={self.modname!r}, dut={self.dut!s})>"

    def __repr__(self):
        modref = self.get_modref()
        return f"<{modref}(inst={self.inst!r}, libname={self.libname!r}, modname={self.modname!r}, dut={self.dut!r})>"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Documentation Container.

The documentation container unifies and eases the handling of documentation strings.
Especially [doc_from_type()][ucdp.docutil.doc_from_type] serves the standard approach,
to create doc for a type related instance.
"""

from .object import LightObject


class Doc(LightObject):
    """
    Documentation Container.

    Documentation is always about a title, a long description and an optional source code comment.
    [Doc][ucdp.doc.Doc] carries all 3 of them.

    Attributes:
        title: Full Spoken Name.
        descr: Documentation Description.
        comment: Source Code Comment. Default is 'title'

    ??? Example "Dictorary Examples"
        Basics:

            >>> from tabulate import tabulate
            >>> import ucdp as u
            >>> docs = (
            ...     u.Doc(),
            ...     u.Doc(title='title'),
            ...     u.Doc(title='title', comment=None),
            ...     u.Doc(descr='descr'),
            ...     u.Doc(comment='comment')
            ... )
            >>> print(tabulate([(doc, doc.title, doc.descr, doc.comment, doc.comment_or_title) for doc in docs],
            ...                headers=("Doc()", ".title", ".descr", ".comment", ".comment_or_title")))
            Doc()                   .title    .descr    .comment    .comment_or_title
            ----------------------  --------  --------  ----------  -------------------
            Doc()
            Doc(title='title')      title                           title
            Doc(title='title')      title                           title
            Doc(descr='descr')                descr
            Doc(comment='comment')                      comment     comment

        Documentation instances are singleton and share the same memory:

            >>> Doc(title='title') is Doc(title='title')
            True
    """

    title: str | None = None
    """
    Full Spoken Name.

    Identifier are often appreviations.
    The ``title`` should contain the full spoken name.

    A signal ``amp_gain`` should have the title ``Amplifier Gain``.
    """

    descr: str | None = None
    """
    Documentation Description.

    The ``descr`` can contain any multiline **user** documentation.
    """

    comment: str | None = None
    """
    Source Code Comment.

    Source code should be commented.
    The ``comment`` can contain any developer / **non-user** documentation.
    Anything useful developer information.
    """

    @property
    def comment_or_title(self):
        """Return `comment` if set, otherwise `title`."""
        return self.comment or self.title



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Overview."""

import anytree

from .iterutil import namefilter


class OverviewNode(anytree.NodeMixin):
    """Overview Node."""

    def __init__(self, title, overview, children, tags):
        super().__init__()
        self.title = title
        self.overview = overview
        self.children = children
        self.tags = tags


def get_overview_tree(mod, minimal=False, tags=None) -> OverviewNode | None:
    """Determine Overview Tree."""
    if minimal:
        if tags:
            # minimal + tags
            tagfilter = namefilter(tags)

            def filter_(node) -> bool:
                return bool(node.tags) and any(tagfilter(tag) for tag in node.tags) and bool(node.get_overview())
        else:
            # minimal
            def filter_(node) -> bool:
                return bool(node.get_overview())
    elif tags:
        # tags
        tagfilter = namefilter(tags)

        def filter_(node) -> bool:
            return bool(node.tags) and any(tagfilter(tag) for tag in node.tags)
    else:

        def filter_(node) -> bool:
            return True

    nodes = tuple(_iter_overview_nodes([mod], filter_))

    if nodes:
        return nodes[0]
    return None


def _iter_overview_nodes(mods, filter_):
    for mod in mods:
        overview = None
        if filter_(mod):
            overview = mod.get_overview()

        children = tuple(_iter_overview_nodes(mod.insts, filter_))

        if overview is not None or children:
            yield OverviewNode(f"{mod.name}  {mod}", overview, children, mod.tags)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Module Utilities.

"""

from collections.abc import Iterator
from inspect import getfile
from pathlib import Path
from typing import Any

from caseconverter import snakecase

from .baseclassinfo import BaseClassInfo, get_baseclassinfos


def get_modbaseinfos(cls_or_inst: Any) -> Iterator[BaseClassInfo]:
    """
    Get Base Classes of `mod`.

    Args:
        cls_or_inst: Class or Instance
    """
    for baseclassinfo in get_baseclassinfos(cls_or_inst):
        if baseclassinfo.libname == "ucdp":
            break
        yield baseclassinfo


def is_tb_from_modname(modname: str) -> bool:
    """
    Determine if module is a testbench component by checking the name.

    Args:
        modname: Module Name
    """
    return modname.endswith("_tb") or "_tb_" in modname


def get_modname(cls):
    """Module Name."""
    return snakecase(cls.__name__.removesuffix("Mod"))


def get_topmodname(cls):
    """Top Module Name."""
    modbaseinfos = tuple(get_modbaseinfos(cls))
    clsname = modbaseinfos[-1].clsname
    return snakecase(clsname.removesuffix("Mod"))


def get_libname(cls) -> str:
    """Module Library Name."""
    return Path(getfile(cls)).parts[-2]


def get_libpath(cls) -> Path:
    """Module Library Path."""
    return Path(getfile(cls)).parent


def get_file(cls, basedir: Path | None = None) -> Path:
    """Get File Path."""
    path = Path(getfile(cls))
    basedir = basedir or Path()
    try:
        return path.relative_to(basedir)
    except ValueError:
        return Path(*path.parts[-2:])



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Base Class Information Tool.
"""

import re
from collections.abc import Iterator
from inspect import getmro, isclass
from typing import Any

from .object import LightObject

_RE_INFO = re.compile(r"^<class '(?P<libname>.+)\.(?P<modname>.+)\.(?P<clsname>.+)'>$")


class BaseClassInfo(LightObject):
    """Base Class Information.

    Attributes:
        cls: Class.
        libname: Python Library Name.
        modname: Python Module Name.
        clsname: Python Class Name.
    """

    cls: Any
    libname: str
    modname: str
    clsname: str


def get_baseclassinfos(cls_or_inst) -> Iterator[BaseClassInfo]:
    """Base Class Info."""
    if not isclass(cls_or_inst):
        cls_or_inst = cls_or_inst.__class__
    for mro in getmro(cls_or_inst):
        mat = _RE_INFO.match(str(mro))
        if not mat:
            break
        yield BaseClassInfo(cls=mro, **mat.groupdict())



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Assigns."""

import re
from pathlib import Path

from pytest import fixture, raises
from test2ref import assert_refdata

import ucdp as u


class ModeType(u.AEnumType):
    """Mode."""

    keytype: u.UintType = u.UintType(2)

    def _build(self):
        self._add(0, "add")
        self._add(1, "sub")
        self._add(2, "max")


class StructType(u.AStructType):
    """My."""

    comment: str = "Mode"

    def _build(self):
        self._add("mode", ModeType())
        self._add("send", u.ArrayType(u.UintType(8), 3))
        self._add("return", u.UintType(4), u.BWD)


class MyType(u.AStructType):
    """A Complex Type."""

    def _build(self):
        self._add("my0", StructType())
        self._add("my1", StructType(), u.BWD)
        self._add("uint", u.UintType(3))


@fixture
def top() -> u.Idents:
    """Top-Identifier."""
    return u.Idents(
        [
            u.Port(MyType(), "port_i"),
            u.Port(MyType(), "port_o"),
            u.Port(MyType(), "other_i"),
            u.Port(MyType(), "other_o"),
            u.Port(StructType(), "struct_i"),
            u.Port(StructType(), "struct_o"),
            u.Signal(StructType(), "struct_s"),
            u.Signal(MyType(), "sig_s"),
            u.Port(u.UintType(8), "data_i"),
            u.Port(u.UintType(8), "data_o"),
            u.Signal(u.UintType(8), "data0_s"),
            u.Signal(u.UintType(8), "data1_s"),
        ]
    )


@fixture
def sub() -> u.Idents:
    """Sub-Identifier."""
    return u.Idents(
        [
            u.Port(MyType(), "sub_i"),
            u.Port(MyType(), "sub_o"),
            u.Port(u.UintType(8), "data_i"),
            u.Port(u.UintType(8), "data_o"),
        ]
    )


def _dump_assigns(assigns: u.Assigns, path: Path, name: str = "assigns", full: bool = False):
    filepath = path / f"{name}.txt"
    with filepath.open("w") as file:
        if not full:
            file.write("SET-ONLY\n")
        for assign in assigns:
            if full or assign.source is not None:
                file.write(f"ASSIGN: {assign}\n")
        if assigns.drivers:
            for dname, value in assigns.drivers:
                file.write(f"DRIVER: {dname}: {value}\n")


def test_basic(top):
    """Assign Basics."""
    doc = u.Doc(title="title", descr="descr", comment="comment")
    target = u.Port(u.UintType(8), "a_i", doc=doc, ifdefs=("IFDEF",))
    source = u.Signal(u.UintType(8), "b_s")
    assign = u.Assign(target=target, source=source)

    assert assign.name == "a_i"
    assert assign.type_ == u.UintType(8)
    assert assign.doc is doc
    assert assign.direction == u.IN
    assert assign.ifdefs == ("IFDEF",)


def test_assign_empty(top, tmp_path):
    """Empty Assigns."""
    assigns = u.Assigns(targets=top, sources=top)
    _dump_assigns(assigns, tmp_path, full=True)
    assert_refdata(test_assign_empty, tmp_path)


def test_assign_empty_inst(top, tmp_path):
    """Empty Assigns for inst."""
    assigns = u.Assigns(targets=top, sources=top, inst=True)
    _dump_assigns(assigns, tmp_path, full=True)
    assert_refdata(test_assign_empty_inst, tmp_path)


def test_reassign(top, tmp_path):
    """Assigns."""
    assigns = u.Assigns(targets=top, sources=top)
    assigns.set(top["port_o"], top["other_i"])

    msg = "'port_o' already assigned to 'other_i'"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(top["port_o"], top["port_i"])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_reassign, tmp_path)


def test_direction_error(top, tmp_path):
    """Assigns."""
    assigns = u.Assigns(targets=top, sources=top)

    msg = "Cannot drive 'other_i' by 'port_i'"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(top["other_i"], top["port_i"])

    assigns.set(top["other_o"], top["port_o"])
    assigns.set(top["port_i"], top["port_o"])
    assigns.set(top["data0_s"], top["data1_s"])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_direction_error, tmp_path)


def test_inst_direction_error(top, tmp_path):
    """Assigns."""
    assigns = u.Assigns(targets=top, inst=True)

    msg = "Cannot drive 'other_o' by 'port_i'"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(top["other_o"], top["port_i"])

    assigns.set(top["other_o"], top["port_o"])
    assigns.set(top["other_i"], top["port_o"])
    assigns.set(top["data0_s"], top["data1_s"])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_inst_direction_error, tmp_path)


def test_drivers(top, tmp_path):
    """Multiple Drivers."""
    drivers = u.Drivers()
    assigns0 = u.Assigns(targets=top, sources=top, drivers=drivers)
    assigns0.set(top["port_o"], top["other_i"])
    assigns1 = u.Assigns(targets=top, sources=top, drivers=drivers)

    msg = "'other_i' already driven by 'other_i'"
    with raises(u.MultipleDriverError, match=re.escape(msg)):
        assigns1.set(top["port_o"], top["other_i"])

    _dump_assigns(assigns0, tmp_path, name="assign0")
    _dump_assigns(assigns1, tmp_path, name="assign1")
    assert_refdata(test_drivers, tmp_path)


def test_lock(top, tmp_path):
    """Lock."""
    assigns = u.Assigns(targets=top, sources=top)
    assert assigns.is_locked is False

    assigns.lock()

    assert assigns.is_locked is True

    msg = "Cannot set 'other_i' to 'port_o'"
    with raises(u.LockError, match=re.escape(msg)):
        assigns.set(top["port_o"], top["other_i"])

    msg = "Cannot set default 'other_i' to 'port_o'"
    with raises(u.LockError, match=re.escape(msg)):
        assigns.set_default(top["port_o"], top["other_i"])

    msg = "Assigns are already locked. Cannot lock again."
    with raises(u.LockError, match=re.escape(msg)):
        assigns.lock()

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_lock, tmp_path)


def test_assign(top, tmp_path):
    """Test Assigns."""
    assigns = u.Assigns(targets=top)

    # valid assignment
    assigns.set(top["port_o"], top["port_i"])

    # re-assignement
    msg = "'port_o' already assigned to 'port_i'"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(top["port_o"], top["other_i"])

    # assign
    msg = "Cannot assign 'data_o' of type UintType(8) to 'port_i' of type MyType()."
    with raises(TypeError, match=re.escape(msg)):
        assigns.set(top["port_i"], top["data_o"])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_assign, tmp_path)


def test_concat(tmp_path):
    """Test Concatenation."""
    a_i = u.Port(u.UintType(4), "a_i")
    b_i = u.Port(u.UintType(4), "b_i")
    c_o = u.Port(u.UintType(8), "c_o")
    idents = u.Idents([a_i, b_i, c_o])
    assigns = u.Assigns(targets=idents)

    assigns.set(c_o, u.ConcatExpr((a_i, b_i)))
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_concat, tmp_path)


def test_const(top, tmp_path):
    """Type Error."""
    assigns = u.Assigns(targets=top)
    constscal = u.Const(u.UintType(8), "scal")
    myconst = u.Const(MyType(), "const_c")

    msg = "Target const_c is not a Signal, Port or Slice of them"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(myconst, top["port_i"])

    assigns.set(top["data_o"], constscal)

    msg = "Target const_my0_return_c is not a Signal or Port"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(top["port_o"], myconst)

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_const, tmp_path)


def test_assign_slice(tmp_path):
    """Test Assign slice."""
    a_i = u.Port(u.UintType(4), "a_i")
    b_i = u.Port(u.UintType(4), "b_i")
    c_i = u.Port(u.UintType(4), "c_i")
    c_o = u.Port(u.UintType(10, default=0x1A1), "c_o")
    idents = u.Idents([a_i, b_i, c_o])
    drivers = u.Drivers()
    assigns = u.Assigns(targets=idents, drivers=drivers)

    assigns.set(c_o[3:1], a_i[2:0])
    assigns.set(c_o[6:5], b_i[2:1])

    msg = "Slice 6:5 is already taken by SliceOp(Port(UintType(4), 'b_i', direction=IN), Slice('2:1'))"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(c_o[6:5], c_i[2:1])

    msg = "Cannot slice bit(s) 11 from UintType(10, default=417)"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(c_o[11], b_i[3])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_assign_slice, tmp_path)


def test_assign_slice_inst(tmp_path):
    """Test Assign slice."""
    a_i = u.Port(u.UintType(4), "a_i")
    b_i = u.Port(u.UintType(4), "b_i")
    c_i = u.Port(u.UintType(4), "c_i")
    c_o = u.Port(u.UintType(10, default=0x1A1), "c_o")
    idents = u.Idents([a_i, b_i, c_i, c_o])
    drivers = u.Drivers()
    assigns = u.Assigns(targets=idents, drivers=drivers, inst=True)

    assigns.set(c_o[3:1], a_i[2:0])
    assigns.set(c_o[6:5], b_i[2:1])

    msg = "Slice 6:5 is already taken by SliceOp(Port(UintType(4), 'b_i', direction=IN), Slice('2:1'))"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(c_o[6:5], c_i[2:1])

    msg = "Cannot slice bit(s) 11 from UintType(10, default=417)"
    with raises(ValueError, match=re.escape(msg)):
        assigns.set(c_o[11], b_i[3])

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_assign_slice_inst, tmp_path)


def test_top_in(top, tmp_path):
    """Top - Input."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=top, drivers=drivers)
    assigns.set(top["port_i"], top["port_o"])
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_in, tmp_path)


def test_top_out(top, tmp_path):
    """Top - Output."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=top, drivers=drivers)
    assigns.set(top["port_o"], top["port_i"])
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_out, tmp_path)


def test_top_inst_in(top, sub, tmp_path):
    """Top with Sub - Input."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=sub, sources=top, inst=True, drivers=drivers)
    assigns.set(sub["sub_i"], top["port_i"])
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_inst_in, tmp_path)


def test_top_inst_out(top, sub, tmp_path):
    """Top with Sub - Output."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=sub, sources=top, inst=True, drivers=drivers)
    assigns.set(sub["sub_o"], top["port_o"])
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_inst_out, tmp_path)


def test_top_inst_in_note(top, sub, tmp_path):
    """Top with Sub - Input."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=sub, sources=top, inst=True, drivers=drivers)
    assigns.set(sub["sub_my1_i"], u.TODO)
    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_inst_in_note, tmp_path)


def test_top_inst_in_const(top, sub, tmp_path):
    """Top with Sub - Input."""
    drivers = u.Drivers()
    assigns = u.Assigns(targets=sub, sources=top, inst=True, drivers=drivers)
    assigns.set(sub["data_i"], u.const("8h2"))

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_top_inst_in_const, tmp_path)


def test_assign_cast(tmp_path):
    """Test Casting."""

    class MyEnumType(u.AEnumType):
        keytype: u.UintType = u.UintType(4)

        def _build(self):
            self._add(u.AUTO, "a")
            self._add(u.AUTO, "b")
            self._add(u.AUTO, "c")

    targets = u.Idents(
        [
            u.Port(u.UintType(4), "data_i"),
            u.Port(MyEnumType(), "a_o"),
            u.Port(MyEnumType(), "b_o"),
            u.Port(MyEnumType(), "c_o"),
        ]
    )
    assigns = u.Assigns(targets=targets)

    msg = "MyEnumType(). Try to cast."
    with raises(TypeError, match=re.escape(msg)):
        assigns.set(targets["a_o"], targets["data_i"])

    assigns.set(targets["b_o"], targets["data_i"], cast=None)

    _dump_assigns(assigns, tmp_path)
    assert_refdata(test_assign_cast, tmp_path)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Utilities."""

import os
import sys
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from functools import lru_cache
from inspect import getfile
from pathlib import Path

from .logging import LOGGER


def get_paths() -> tuple[Path, ...]:
    """Determine Paths from Environment variable 'UCDP_PATH'."""
    return tuple(Path(item.strip()) for item in os.environ.get("UCDP_PATH", "").split())


@contextmanager
def extend_sys_path(paths: Iterable[Path] | None = None, use_env_default: bool = False):
    """Context with extended sys.path.

    Keyword Args:
        paths: Paths.
        use_env_default: Use UCDP_PATH as default if paths are unset.
    """
    if paths is None and use_env_default:
        paths = get_paths()
    globpaths = []
    for path in paths or []:
        globpaths.extend(glob(path))
    pathstrs = [str(path) for path in globpaths]
    LOGGER.debug("paths=%r", pathstrs)
    if pathstrs:
        orig = sys.path
        sys.path = [*sys.path, *pathstrs]
        yield
        sys.path = orig
    else:
        yield


def get_copyright(obj: Path | object) -> str:
    """Determine from Source Code of ``obj``."""
    if isinstance(obj, Path):
        path = obj
    else:
        path = Path(getfile(obj.__class__))
    return _get_copyright(path)


@lru_cache
def _get_copyright(path: Path) -> str:
    lines = []
    with path.open(encoding="utf-8") as file:
        for line in file:
            if line.startswith("#"):
                lines.append(line[1:])
            else:
                break
    return "".join(lines)


def get_maxworkers() -> int | None:
    """Maximum Number of Workers."""
    try:
        return int(os.environ["UCDP_MAXWORKERS"])
    except (KeyError, ValueError):
        return None


def guess_path(arg: str) -> Path | None:
    """Return Path if arg seems to be a file path."""
    path = Path(arg)
    if path.exists() or len(path.parts) > 1:
        return path
    return None


def glob(path: Path) -> Iterator[Path]:
    """Glob `path`."""
    assert isinstance(path, Path), path
    path = path.absolute()
    pathstr = str(path)
    if "*" in pathstr or "?" in pathstr or "[" in pathstr:
        root = Path(path.parts[0])
        pattern = str(path.relative_to(root))
        yield from sorted(root.glob(pattern))
    else:
        yield path



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import re

from pytest import raises
from test2ref import assert_refdata

import ucdp as u


class MuxMod(u.AMod):
    """Module using Multiplexer."""

    def _build(self):
        sel = self.add_signal(u.UintType(3), "sel_s")
        self.add_port(u.UintType(4), "a0_i")
        b0 = self.add_port(u.UintType(4), "b0_i")
        self.add_port(u.UintType(4), "c0_i")
        q0 = self.add_port(u.UintType(4), "q0_o")

        self.add_port(u.UintType(8), "a1_i")
        self.add_port(u.UintType(8), "b1_i")
        self.add_port(u.UintType(8), "c1_i")
        self.add_port(u.UintType(8), "q1_o")

        self.add_port(u.UintType(8), "q2_o")

        mux = self.add_mux("main", title="title", descr="descr", comment="comment")

        mux.set_default("q0_o", u.UintType(4, default=8))
        mux.set("sel_s", "3h1", "q0_o", "a0_i")
        mux.set(sel, "3h2", q0, b0)
        mux.set(sel, "3h4", q0, "c0_i")

        mux.set("sel_s", "3h0", "q1_o", "a1_i")
        mux.set("sel_s", "3h1", "q1_o", "b1_i")
        mux.set_default("q1_o", "c1_i")

        mux = self.add_mux("slim")
        mux.set("sel_s", "3h1", "q2_o", "a1_i")

        mux = self.add_mux("empty")


def test_mux(tmp_path, capsys):
    """Multiplexer."""
    mod = MuxMod()

    mainmux = mod.get_mux("main")
    assert mainmux.name == "main"
    assert mainmux.doc == u.Doc(title="title", descr="descr", comment="comment")
    assert mod.get_mux(mainmux) is mainmux

    emptymux = mod.get_mux("empty")
    assert emptymux.name == "empty"
    assert emptymux.doc == u.Doc()

    msg = "'missing'. Known are 'main', 'slim' and 'empty'."
    with raises(ValueError, match=re.escape(msg)):
        mod.get_mux("missing")

    _print_muxes(mod)
    assert_refdata(test_mux, tmp_path, capsys=capsys)


def _print_muxes(mod: u.BaseMod):
    rslvr = u.ExprResolver()
    for mux in mod.muxes:
        for selexpr, conds in mux:
            print(mux.name, mux.sels)
            sel = rslvr(selexpr)
            for default in mux.defaults():
                print(default)
            for condexpr, assigns in conds.items():
                cond = rslvr(condexpr)
                print(f"  {sel}=={cond}:")
                for assign in assigns:
                    print(f"    {assign}")
        for name, driver in mux.drivers:
            print(f"  {name}: {driver}")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Flip Flop."""

from .assigns import Assigns
from .expr import BoolOp
from .object import model_validator
from .signal import BaseSignal
from .typeclkrst import ClkType, RstAnType


class FlipFlop(Assigns):
    """FlipFlop.

    Attributes:
        clk: Clock
        rst_an: Reset (low active)
        rst: Reset
        ena: Enable

    """

    clk: BaseSignal
    rst_an: BaseSignal
    rst: BoolOp | None = None
    ena: BoolOp | None = None

    @model_validator(mode="after")
    def __post_init(self) -> "FlipFlop":
        if not isinstance(self.clk.type_, ClkType):
            raise ValueError("clk is not of ClkType")
        if not isinstance(self.rst_an.type_, RstAnType):
            raise ValueError("rst_an is not of RstAnType")
        return self



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Scalar Type Testing.

* :any:`IntegerType`
* :any:`BitType`
* :any:`BoolType`
* :any:`RailType`
* :any:`UintType`
* :any:`SintType`

"""

from pytest import raises

import ucdp as u


def test_bit():
    """Bit."""
    assert u.BitType() is u.BitType()
    assert u.BitType() is not u.BitType(default=0)
    assert u.BitType() is not u.BitType(default=1)
    with raises(u.ValidationError):
        u.BitType(width=3)


def test_integer():
    """Integer."""
    var0 = u.IntegerType()
    assert var0.default == 0
    assert var0.width == 32
    assert repr(var0) == "IntegerType()"
    with raises(u.ValidationError):
        var0.default = 0xF
    assert var0.default == 0

    assert 1 in var0

    var1 = u.IntegerType(default=8)
    assert var1.default == 8
    assert var1.width == 32
    assert repr(var1) == "IntegerType(default=8)"

    var2 = u.IntegerType()
    assert var2.default == 0
    assert var2.width == 32
    assert repr(var2) == "IntegerType()"

    with raises(u.ValidationError):
        u.IntegerType(width=4)

    assert var0 is not var1
    assert var0 is var2
    assert var0 != var1
    assert var0 == var2

    with raises(ValueError):
        u.IntegerType(default="safe")

    assert u.IntegerType() is u.IntegerType()
    assert u.IntegerType() is not u.IntegerType(default=0)
    assert u.IntegerType() is not u.IntegerType(default=1)


def test_rail():
    """Rail."""
    var0 = u.RailType()
    assert var0.default is None
    assert var0.width == 1
    assert repr(var0) == "RailType()"
    with raises(u.ValidationError):
        var0.default = 0xF
    assert var0.default is None

    assert 1 in var0

    var1 = u.RailType(default=1)
    assert var1.default == 1
    assert var1.width == 1
    assert repr(var1) == "RailType(default=1)"

    var2 = u.RailType()
    assert var2.default is None
    assert var2.width == 1
    assert repr(var2) == "RailType()"

    with raises(u.ValidationError):
        u.RailType(width=4)

    assert var0 is not var1
    assert var0 is var2
    assert var0 != var1
    assert var0 == var2

    with raises(ValueError):
        u.RailType(default="safe")

    assert u.RailType() is u.RailType()
    assert u.RailType() is not u.RailType(default=0)
    assert u.RailType() is not u.RailType(default=1)


def test_uint():
    """Uint Vector."""
    var0 = u.UintType(12)
    assert var0.default == 0
    assert var0.width == 12
    assert var0.bits == 12
    assert repr(var0) == "UintType(12)"

    assert -1 not in var0
    assert 1 in var0
    assert 2**12 not in var0

    var1 = u.UintType(12, default=8)
    assert var1.default == 8
    assert var1.width == 12
    assert var1.right == 0
    assert var1.slice_ == u.Slice("11:0")
    assert repr(var1) == "UintType(12, default=8)"

    assert repr(var1.new(width=4)) == "UintType(4, default=8)"

    with raises(u.ValidationError):
        var1.default = 1

    assert var0 is not var1
    assert var0 != var1

    with raises(ValueError):
        u.UintType(12, default="abc")

    var1 = u.UintType(12, default=8, right=4)
    assert var1.default == 8
    assert var1.width == 12
    assert var1.right == 4
    assert var1.slice_ == u.Slice("15:4")
    assert repr(var1) == "UintType(12, default=8, right=4)"


def test_sint():
    """Sint Vector."""
    var0 = u.SintType(12)
    assert var0.default == 0
    assert var0.width == 12
    assert var0.bits == 12
    assert repr(var0) == "SintType(12)"

    assert -1 in var0
    assert 1 in var0
    assert -(2**11) in var0
    assert 2**11 - 1 in var0
    assert 2**11 not in var0

    var1 = u.SintType(12, default=8)
    assert var1.default == 8
    assert var1.width == 12
    assert var1.right == 0
    assert var1.slice_ == u.Slice("11:0")
    assert repr(var1) == "SintType(12, default=8)"

    assert repr(var1.new(width=5)) == "SintType(5, default=8)"
    with raises(u.ValidationError):
        var1.new(width=4)

    with raises(u.ValidationError):
        var1.default = 1

    assert var0 is not var1
    assert var0 != var1

    with raises(ValueError):
        u.SintType(12, default="abc")

    var1 = u.SintType(12, default=8, right=4)
    assert var1.default == 8
    assert var1.width == 12
    assert var1.right == 4
    assert var1.slice_ == u.Slice("15:4")
    assert repr(var1) == "SintType(12, default=8, right=4)"


def test_doc():
    """Documentation."""
    assert u.UintType(8).doc == u.Doc()

    class MyUintType(u.UintType):
        title: str = "mytitle"
        descr: str = "mydescr"
        comment: str = "mycomment"

        def __init__(self):
            super().__init__(8)

    assert MyUintType().doc == u.Doc(title="mytitle", descr="mydescr", comment="mycomment")


class MyUintType(u.UintType):
    """Example Uint."""

    def __init__(self, width=42):
        super().__init__(width=width)


def test_fixed_vector_size():
    """Uint with fixed size."""
    assert MyUintType().width == 42



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Scalar Types aka native Types."""

from typing import Any

from .object import Light
from .typebase import BaseType


class FloatType(BaseType, Light):
    """
    Native Floating Point Number.

    Example:
    >>> import ucdp as u
    >>> example = u.FloatType()
    >>> example
    FloatType()
    """

    default: Any = 0

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`FloatType`.

        >>> import ucdp as u
        >>> u.FloatType().is_connectable(u.FloatType())
        True

        A connection to other types is forbidden.

        >>> u.FloatType().is_connectable(u.UintType(1))
        False
        """
        return isinstance(other, FloatType)

    def __getitem__(self, slice_):
        """
        Return Sliced Variant.
        """
        raise ValueError("Slicing is not allowed floating point numbers")


class DoubleType(BaseType, Light):
    """
    Native Double Precision Floating Point Number.

    Example:
    >>> import ucdp as u
    >>> example = u.DoubleType()
    >>> example
    DoubleType()
    """

    default: Any = 0

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`DoubleType`.

        >>> import ucdp as u
        >>> u.DoubleType().is_connectable(u.DoubleType())
        True

        A connection to other types is forbidden.

        >>> u.DoubleType().is_connectable(u.UintType(1))
        False
        """
        return isinstance(other, DoubleType)

    def __getitem__(self, slice_):
        """
        Return Sliced Variant.
        """
        raise ValueError("Slicing is not allowed on floating point numbers")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Path Utilities."""

import os
import re
from pathlib import Path
from unittest import mock

from contextlib_chdir import chdir
from pytest import raises

import ucdp as u


def test_improved_glob(tmp_path):
    """Test improved_glob."""
    sub = tmp_path / "sub"
    sub.mkdir()
    song = tmp_path / "song.txt"
    song.touch()
    song2 = tmp_path / "song2.txt"
    song2.touch()
    subsong = sub / "song.txt"
    subsong.touch()
    nothing = tmp_path / "nothing.txt"

    # existing
    assert tuple(u.improved_glob("song.txt", basedir=tmp_path)) == (song,)
    assert tuple(u.improved_glob("sub/song.txt", basedir=tmp_path)) == (subsong,)

    # not existing one - do not clear
    assert tuple(u.improved_glob("song.txt")) == (Path("song.txt"),)
    assert tuple(u.improved_glob("nothing.txt", basedir=tmp_path)) == (nothing,)

    # glob
    assert tuple(u.improved_glob("song*.txt", basedir=tmp_path)) == (song, song2)

    # recursive
    assert tuple(u.improved_glob("**/*.txt", basedir=tmp_path)) == (song, song2, subsong)


def test_improved_glob_env(tmp_path):
    """Test improved_glob with Env."""
    sub = tmp_path / "sub"
    sub.mkdir()
    song = tmp_path / "song.txt"
    song.touch()
    song2 = tmp_path / "song2.txt"
    song2.touch()
    subsong = sub / "song.txt"
    subsong.touch()

    # existing songs
    with mock.patch.dict(os.environ, {"SUB": str(sub), "TMP": str(tmp_path)}):
        assert tuple(u.improved_glob("${SUB}/song.txt")) == (Path("${SUB}/song.txt"),)

        # glob
        assert tuple(u.improved_glob("${SUB}/song*.txt")) == (Path("${SUB}/song.txt"),)
        assert tuple(u.improved_glob("${SUB}/../song*.txt")) == (
            Path("${SUB}/../song.txt"),
            Path("${SUB}/../song2.txt"),
        )
        assert tuple(u.improved_glob("${TMP}/song*.txt")) == (
            Path("${TMP}/song.txt"),
            Path("${TMP}/song2.txt"),
        )

        # recursive
        assert tuple(u.improved_glob("${TMP}/**/*.txt")) == (
            Path("${TMP}/song.txt"),
            Path("${TMP}/song2.txt"),
            Path("${TMP}/sub/song.txt"),
        )


def test_improved_resolve(tmp_path):
    """Improved Resolve."""
    sub = tmp_path / "sub"
    sub.mkdir()
    song = tmp_path / "song.txt"
    song.touch()
    song2 = tmp_path / "song2.txt"
    song2.touch()
    subsong = sub / "song.txt"
    subsong.touch()
    nothing = tmp_path / "nothing.txt"

    assert u.improved_resolve(Path("song.txt"), basedir=tmp_path) == song
    assert u.improved_resolve(Path("song.txt"), basedir=tmp_path, strict=True) == song

    assert u.improved_resolve(Path("nothing.txt"), basedir=tmp_path) == nothing
    with raises(FileNotFoundError):
        u.improved_resolve(Path("nothing.txt"), basedir=tmp_path, strict=True)

    with chdir(sub):
        assert u.improved_resolve(Path("song.txt")) == subsong


def test_improved_resolve_env(tmp_path):
    """Improved Resolve with environment variables."""
    env1 = tmp_path / "env1"
    env1.mkdir()
    env2 = tmp_path / "env2"
    song1 = env1 / "song.txt"
    song1.touch()
    song2 = env2 / "song.txt"

    with mock.patch.dict(os.environ, {"ENV1": str(env1), "ENV2": str(env2)}):
        # defined and existing env path
        song1file = Path("${ENV1}") / "song.txt"
        assert u.improved_resolve(Path("${ENV1}/song.txt")) == song1file
        assert u.improved_resolve(Path("${ENV1}/song.txt"), replace_envvars=True) == song1
        assert u.improved_resolve(Path("${ENV1}/song.txt"), replace_envvars=True, strict=True) == song1
        with raises(FileNotFoundError, match=re.escape(str(song1file))):
            u.improved_resolve(Path("${ENV1}/song.txt"), strict=True)

        # defined and not existing env path
        song2file = Path("${ENV2}") / "song.txt"
        assert u.improved_resolve(Path("${ENV2}/song.txt")) == song2file
        assert u.improved_resolve(Path("${ENV2}/song.txt"), replace_envvars=True) == song2
        with raises(FileNotFoundError):
            u.improved_resolve(Path("${ENV2}/song.txt"), replace_envvars=True, strict=True)
        with raises(FileNotFoundError):
            u.improved_resolve(Path("${ENV2}/song.txt"), strict=True)

        # not defined env path
        song3file = Path("${ENV3}") / "song.txt"
        assert u.improved_resolve(Path("${ENV3}/song.txt")) == song3file
        assert (
            u.improved_resolve(Path("${ENV3}/song.txt"), replace_envvars=True)
            == Path().resolve() / Path("${ENV3}") / "song.txt"
        )
        with raises(FileNotFoundError, match=re.escape("${ENV3}")):
            u.improved_resolve(Path("${ENV3}/song.txt"), replace_envvars=True, strict=True)
        with raises(FileNotFoundError, match=re.escape(str(song3file))):
            u.improved_resolve(Path("${ENV3}/song.txt"), strict=True)


def test_use_envvars(tmp_path):
    """Test Use Envvars."""
    env1 = tmp_path / "env1"
    env2 = tmp_path / "env2"
    env2sub = env2 / "sub"
    env3 = tmp_path / "env3"
    other = tmp_path / "other"
    envvarnames = ("ENV1", "env2sub", "env2", "env3")
    with mock.patch.dict(os.environ, {"ENV1": str(env1), "env2sub": str(env2sub), "env2": str(env2)}):
        assert u.use_envvars(other, envvarnames) == other

        # defined and existing env path
        assert u.use_envvars(env1, envvarnames) == Path("${ENV1}")
        assert u.use_envvars(env1 / "file.txt", envvarnames) == Path("${ENV1}") / "file.txt"

        # defined and not existing env path
        assert u.use_envvars(env2, envvarnames) == Path("${env2}")
        assert u.use_envvars(env2sub, envvarnames) == Path("${env2sub}")

        # not defined
        assert u.use_envvars(env3, envvarnames) == env3

        assert u.use_envvars(Path(), envvarnames) == Path()


def test_startswith_envvar(tmp_path):
    """Test startswith_envvar."""
    env1 = tmp_path / "env1"
    env1.mkdir()
    env2 = tmp_path / "env2"
    with mock.patch.dict(os.environ, {"env1": str(env1), "env2": str(env2)}):
        assert u.startswith_envvar(env1) == (None, env1)

        # defined and existing env path
        assert u.startswith_envvar(Path("$env1")) == ("env1", Path())
        assert u.startswith_envvar(Path("$env1") / "sub") == ("env1", Path("sub"))
        assert u.startswith_envvar(Path("$env1") / "sub", strict=True) == ("env1", Path("sub"))
        assert u.startswith_envvar(Path("$env1") / "sub", barename=True) == ("env1", Path("sub"))

        assert u.startswith_envvar(Path("${env1}")) == ("{env1}", Path())
        assert u.startswith_envvar(Path("${env1}") / "sub") == ("{env1}", Path("sub"))
        assert u.startswith_envvar(Path("${env1}"), barename=True) == ("env1", Path())
        assert u.startswith_envvar(Path("${env1}") / "sub", barename=True) == ("env1", Path("sub"))

        # defined and not existing env path
        assert u.startswith_envvar(Path("$env2")) == ("env2", Path())
        assert u.startswith_envvar(Path("$env2") / "sub") == ("env2", Path("sub"))
        with raises(FileNotFoundError):
            u.startswith_envvar(Path("$env2") / "sub", strict=True)

        # undefined env path
        assert u.startswith_envvar(Path("$env3")) == ("env3", Path())
        assert u.startswith_envvar(Path("$env3") / "sub") == ("env3", Path("sub"))
        with raises(FileNotFoundError):
            u.startswith_envvar(Path("$env3") / "sub", strict=True)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Expression Parser.
"""

from collections.abc import Iterable
from functools import cached_property
from typing import Any

from ._castingnamespace import CastingNamespace
from .consts import RE_IDENTIFIER
from .exceptions import InvalidExpr
from .expr import (
    _RE_CONST,
    BoolOp,
    ConcatExpr,
    ConstExpr,
    Expr,
    Log2Expr,
    MaximumExpr,
    MinimumExpr,
    Op,
    SliceOp,
    SOp,
    TernaryExpr,
    _parse_const,
)
from .namespace import Namespace
from .note import Note
from .object import Object, computed_field
from .typebase import BaseScalarType, BaseType
from .typescalar import BoolType

Parseable = Expr | str | int | BaseType | list | tuple
Constable = int | str | ConstExpr
Concatable = list | tuple | ConcatExpr
Only = type[Expr] | Iterable[type[Expr]] | type[Note]
Types = type[BaseType] | Iterable[type[BaseType]]


class _Globals(dict):
    def __init__(self, globals: dict, namespace: Namespace | CastingNamespace | None, context: str | None):
        super().__init__(globals)
        self.namespace = namespace
        self.context = context

    def __missing__(self, key):
        if self.namespace:
            try:
                return self.namespace.get_dym(key)
            except ValueError as err:
                raise NameError(f"{self.context}: {err}") from None
        return NameError(key)


class ExprParser(Object):
    """
    ExprParser.

    Attributes:
        namespace (Namespace): Symbol namespace
    """

    namespace: Namespace | CastingNamespace | None = None
    context: str | None = None

    @computed_field
    @cached_property
    def _globals(self) -> _Globals:
        globals_ = {
            # Expressions
            "Op": Op,
            "SOp": SOp,
            "BoolOp": BoolOp,
            "SliceOp": SliceOp,
            "ConstExpr": ConstExpr,
            "ConcatExpr": ConcatExpr,
            "TernaryExpr": TernaryExpr,
            "Log2Expr": Log2Expr,
            "MinimumExpr": MinimumExpr,
            "MaximumExpr": MaximumExpr,
            # Helper
            "const": self.const,
            "concat": self.concat,
            "ternary": self.ternary,
            "log2": self.log2,
            "minimum": self.minimum,
            "maximum": self.maximum,
        }
        return _Globals(globals=globals_, namespace=self.namespace, context=self.context)

    def __call__(self, expr: Parseable, only: Only | None = None, types: Types | None = None) -> Expr:
        """
        Parse Expression.

        This is an alias to `parse`.

        Args:
            expr: Expression

        Keyword Args:
            only: Limit expression to these final element type.
            types: Limit expression type to to these types.
        """
        return self.parse(expr, only=only, types=types)

    def parse_note(self, expr: Parseable | Note, only: Only | None = None, types: Types | None = None) -> Expr | Note:
        """
        Parse Expression or Note.

        Args:
            expr: Expression

        Keyword Args:
            only: Limit expression to these final element type.
            types: Limit expression type to to these types.
        """
        if isinstance(expr, Note):
            self._check(expr, only=only, types=types)
            return expr
        return self.parse(expr, only=only, types=types)

    def parse(self, expr: Parseable, only: Only | None = None, types: Types | None = None) -> Expr:
        """
        Parse Expression.

        Args:
            expr: Expression

        Keyword Args:
            only: Limit expression to these final element type.
            types: Limit expression type to to these types.

        ??? Example "Expression Parser Examples"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> p.parse(10)
                ConstExpr(IntegerType(default=10))
                >>> p.parse('3h3')
                ConstExpr(UintType(3, default=3))
                >>> p.parse('3h3') * p.const(2)
                Op(ConstExpr(UintType(3, default=3)), '*', ConstExpr(IntegerType(default=2)))
                >>> p.parse((10, '10'))
                ConcatExpr((ConstExpr(IntegerType(default=10)), ConstExpr(IntegerType(default=10))))
                >>> p = u.ExprParser(namespace=u.Idents([
                ...     u.Signal(u.UintType(16, default=15), 'uint_s'),
                ...     u.Signal(u.SintType(16, default=-15), 'sint_s'),
                ... ]))
                >>> expr = p.parse('uint_s[2]')
                >>> expr
                SliceOp(Signal(UintType(16, default=15), 'uint_s'), Slice('2'))
                >>> expr = p.parse('uint_s * sint_s[2:1]')
                >>> expr
                Op(Signal(UintType(16, ...), 'uint_s'), '*', SliceOp(Signal(SintType(16, ...), 'sint_s'), Slice('2:1')))
                >>> int(expr)
                0

            A more complex:

                >>> namespace = u.Idents([
                ...     u.Signal(u.UintType(2), 'a_s'),
                ...     u.Signal(u.UintType(4), 'b_s'),
                ...     u.Signal(u.SintType(8), 'c_s'),
                ...     u.Signal(u.SintType(16), 'd_s'),
                ... ])
                >>> p = u.ExprParser(namespace=namespace)
                >>> expr = p.parse("ternary(b_s == const('4h3'), a_s, c_s)")
                >>> expr
                TernaryExpr(BoolOp(Signal(UintType(4), 'b_s'), '==', ..., Signal(SintType(8), 'c_s'))

                Syntax Errors:

                >>> parse("sig_s[2")  # doctest: +SKIP
                Traceback (most recent call last):
                ...
                u.exceptions.InvalidExpr: 'sig_s[2': '[' was never closed (<string>, line 1)
        """
        result: Expr
        if isinstance(expr, Expr):
            result = expr
        elif isinstance(expr, BaseType):
            result = ConstExpr(expr)
        else:
            try:
                if isinstance(expr, (list, tuple)):
                    result = self.concat(expr)
                else:
                    try:
                        result = self.const(expr)
                    except InvalidExpr:
                        result = self._parse(str(expr))
            except NameError as exc:
                raise exc
        self._check(result, only=only, types=types)
        return result

    def _check(self, expr: Expr | Note, only: Only | None, types: Types | None) -> None:
        if only and not isinstance(expr, only):  # type: ignore[arg-type]
            raise ValueError(f"{expr!r} is not a {only}. It is a {type(expr)}") from None
        if types:
            if isinstance(expr, Note):
                raise ValueError(f"{expr!r} does not meet type_ {types}.") from None
            if not isinstance(expr.type_, types):  # type: ignore[arg-type]
                raise ValueError(f"{expr!r} requires type_ {types}. It is a {expr.type_}") from None

    def _parse(self, expr: str) -> Expr:
        if self.namespace:
            # avoid eval call on simple identifiers
            if isinstance(expr, str) and RE_IDENTIFIER.match(expr):
                try:
                    return self.namespace[expr]
                except ValueError:
                    pass
        if _RE_CONST.fullmatch(expr.strip()):
            return _parse_const(expr)
        try:
            globals: dict[str, Any] = self._globals  # type: ignore[assignment]
            return eval(expr, globals)  # noqa: S307
        except TypeError:
            raise InvalidExpr(expr) from None
        except SyntaxError as exc:
            raise InvalidExpr(f"{expr!r}: {exc!s}") from None

    def const(self, value: Constable) -> ConstExpr:
        """
        Parse Constant.

        ??? Example "Parser Example"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> p.const('10')
                ConstExpr(IntegerType(default=10))
                >>> p.const(10)
                ConstExpr(IntegerType(default=10))
                >>> p.const("10'd20")
                ConstExpr(UintType(10, default=20))
                >>> p.const(u.ConstExpr(u.UintType(10, default=20)))
                ConstExpr(UintType(10, default=20))
                >>> p.const("4'h4")
                ConstExpr(UintType(4, default=4))
                >>> p.const("4'sh4")
                ConstExpr(SintType(4, default=4))
                >>> p.const("4'shC")
                ConstExpr(SintType(4, default=-4))
        """
        if isinstance(value, ConstExpr):
            return value
        return _parse_const(value)

    def concat(self, value: Concatable) -> ConcatExpr:
        """
        Parse ConcatExpr.

        ??? Example "Concat Parser Example"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> p.concat((10, "20"))
                ConcatExpr((ConstExpr(IntegerType(default=10)), ConstExpr(IntegerType(default=20))))

                >>> bool(p.concat((10, "20")))
                True
        """
        if isinstance(value, ConcatExpr):
            return value
        return ConcatExpr(tuple(self.parse(item) for item in value))

    def ternary(self, cond: Parseable, one: Parseable, other: Parseable) -> TernaryExpr:
        """
        TernaryExpr Statement.

        ??? Example "Ternary Parser Example"
            Basics:

                >>> import ucdp as u
                >>> cond = u.Signal(u.UintType(2), 'if_s') == u.ConstExpr(u.UintType(2, default=1))
                >>> one = u.Signal(u.UintType(16, default=10), 'one_s')
                >>> other = u.Signal(u.UintType(16, default=20), 'other_s')
                >>> p = u.ExprParser()
                >>> expr = p.ternary(cond, one, other)
                >>> expr
                TernaryExpr(BoolOp(Signal(UintType(2), 'if_s'), '==', ..., Signal(UintType(16, default=20), 'other_s'))
                >>> int(expr)
                20
                >>> expr.type_
                UintType(16, default=10)
        """
        condp: BoolOp = self.parse(cond, only=BoolOp)  # type:ignore[assignment]
        onep = self.parse(one)
        otherp = self.parse(other)
        return TernaryExpr(cond=condp, one=onep, other=otherp)

    def log2(self, expr: Parseable):
        """
        Ceiling Logarithm to base of 2.

        ??? Example "Log2 Parser Example"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> log = p.log2("8'h8")
                >>> log
                Log2Expr(ConstExpr(UintType(8, default=8)))
                >>> int(log)
                3
                >>> p.parse("log2('8h8')")
                Log2Expr(ConstExpr(UintType(8, default=8)))
        """
        return Log2Expr(self.parse(expr))

    def minimum(self, *items):
        """
        Lower value of `one` and `other`.

        ??? Example "Minimum Parser Example"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> val = p.minimum("8'h8", "8'h3")
                >>> val
                MinimumExpr((ConstExpr(UintType(8, default=8)), ConstExpr(UintType(8, default=3))))
                >>> int(val)
                3
                >>> p.parse("minimum('8h8', '8h3')")
                MinimumExpr((ConstExpr(UintType(8, default=8)), ConstExpr(UintType(8, default=3))))
        """
        parsed = tuple(self.parse(item) for item in items)
        return MinimumExpr(parsed)

    def maximum(self, *items):
        """
        Higher value of `one` and `other`.

        ??? Example "Maximum Parser Example"
            Basics:

                >>> import ucdp as u
                >>> p = u.ExprParser()
                >>> val = p.maximum("8'h8", "8'h3")
                >>> val
                MaximumExpr((ConstExpr(UintType(8, default=8)), ConstExpr(UintType(8, default=3))))
                >>> int(val)
                8
                >>> p.parse("maximum('8h8', '8h3')")
                MaximumExpr((ConstExpr(UintType(8, default=8)), ConstExpr(UintType(8, default=3))))
        """
        parsed = tuple(self.parse(item) for item in items)
        return MaximumExpr(parsed)


def cast_booltype(expr):
    """Cast to Boolean."""
    type_ = expr.type_
    if isinstance(type_, BoolType):
        return expr
    if isinstance(type_, BaseScalarType) and int(type_.width) == 1:
        return expr == 1
    raise ValueError(f"{expr} does not result in bool")


_PARSER = ExprParser()
const = _PARSER.const
concat = _PARSER.concat
ternary = _PARSER.ternary
log2 = _PARSER.log2
minimum = _PARSER.minimum
maximum = _PARSER.minimum



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Basic Types.

DOCME

* :any:`Type` - Base Class for all Types
* :any:`AScalarType` - Base Class for all Single Value Types
* :any:`VecType` - Base Class for all vector-mapped Scalars.
* :any:`CompositeType` - Base class for all assembling Types.
"""

from abc import abstractmethod
from typing import Any, ClassVar

from pydantic import model_validator

from .casting import Casting
from .doc import Doc
from .object import Light, Object, PosArgs
from .slices import DOWN, Slice


class BaseType(Object):
    """
    Base Class for all Types.
    """

    title: str | None = None
    descr: str | None = None
    comment: str | None = None

    @property
    def doc(self) -> Doc:
        """Documentation."""
        return Doc(title=self.title, descr=self.descr, comment=self.comment)

    def is_connectable(self, other: "BaseType"):
        """
        Check For Valid Connection To `other`.

        This method has to be overwritten.
        """
        raise NotImplementedError

    def cast(self, other: "BaseType") -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        return None

    @property
    def bits(self):
        """
        Size in Bits.

        This method has to be overwritten.
        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, slice_):
        """
        Return Sliced Variant.
        """


class BaseScalarType(BaseType):
    """
    Base Type For All Native Types.
    """

    default: Any | None = None


class AScalarType(BaseScalarType, Light):
    """
    Abstract Type For All Native Types.
    """

    width: Any = 1

    @model_validator(mode="after")
    def __post_init(self) -> "AScalarType":
        default = self.default
        if default is not None:
            self.check(default, what="default")
        return self

    def check(self, value, what="Value"):
        """
        Check if `value` can be handled by type`.

        >>> import ucdp as u
        >>> atype = u.AScalarType()
        >>> atype.check(1)
        1
        """
        return value  # pragma: no cover

    def encode(self, value, usedefault=False):
        """
        Encode Value.

        >>> import ucdp as u
        >>> atype = u.AScalarType()
        >>> atype.encode(1)
        1
        """
        return int(value)

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.AScalarType().get_hex()
        """
        return

    def __getitem__(self, slice_):
        """
        Return Sliced Variant.

        >>> import ucdp as u
        >>> u.AScalarType()[4]
        Traceback (most recent call last):
            ...
        ValueError: Cannot slice (4) from AScalarType()
        """
        slice_ = Slice.cast(slice_, direction=DOWN)
        raise ValueError(f"Cannot slice ({slice_}) from {self}")

    def __contains__(self, item):
        try:
            if isinstance(item, range):
                items = tuple(item)
                self.check(self.encode(items[0]))
                self.check(self.encode(items[-1]))
            else:
                self.check(self.encode(item))
            return True
        except ValueError:
            return False

    @property
    def bits(self):
        """Size in Bits."""
        return self.width


class AVecType(AScalarType):
    """Base Class for all Vector Types."""

    default: Any = 0
    """Default Value."""

    right: Any = 0
    """Right Bit Position."""

    logic: bool = True
    """Include X and Z states, not just numeric values."""

    _posargs: ClassVar[PosArgs] = ("width",)

    def __init__(self, width, **kwargs):
        super().__init__(width=width, **kwargs)

    @property
    def slice_(self):
        """Slice."""
        return Slice(width=self.width, right=self.right)


class ACompositeType(BaseType):
    """Base Class For All Composite Types."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
enum Type Testing.

"""

from pytest import raises

import ucdp as u


def test_enum():
    """Enum."""

    class MyEnumType(u.AEnumType):
        """Enum."""

        keytype: u.AScalarType = u.UintType(4)

        def _build(self) -> None:
            self._add(u.AUTO, "a", title="title")
            self._add(u.AUTO, "b", descr="descr")
            self._add(4, "a")
            self._add(u.AUTO, "d", comment="comment")
            with raises(ValueError):
                self._add(5, 8)

    enum = MyEnumType()
    with raises(u.LockError):
        enum._add(u.AUTO, 11)
    assert tuple(enum) == tuple(enum.keys())
    assert tuple(enum.keys()) == (0, 1, 4, 5)
    assert tuple(enum.values()) == (
        u.EnumItem(0, "a", doc=u.Doc(title="title")),
        u.EnumItem(1, "b", doc=u.Doc(descr="descr")),
        u.EnumItem(4, "a"),
        u.EnumItem(5, "d", doc=u.Doc(comment="comment")),
    )
    assert [item.doc for item in enum.values()] == [
        u.Doc(title="title"),
        u.Doc(descr="descr"),
        u.Doc(),
        u.Doc(comment="comment"),
    ]

    assert MyEnumType() is MyEnumType()
    assert MyEnumType() == MyEnumType()


def test_inherit_baseenum():
    """:any:`BaseEnumType` must be inherited."""
    with raises(TypeError):
        u.BaseEnumType()


def test_inherit_enum():
    """:any:`EnumType` must be inherited."""
    with raises(TypeError):
        u.AEnumType()


def test_inherit_globalenum():
    """:any:`AGlobalEnumType` must be inherited."""
    with raises(TypeError):
        u.AGlobalEnumType()


def test_inherit_dynamicenum():
    """:any:`DynamicEnumType` must be inherited."""
    u.DynamicEnumType()


def test_globalenum():
    """Global enum."""

    class MyType(u.AGlobalEnumType):
        pass

    one = MyType()
    one.add(0, "a")
    one.add(1, "b")
    assert tuple(one) == (0, 1)

    other = MyType()
    other.add(2, "c")

    assert one is other
    assert one == other
    assert tuple(one) == (0, 1, 2)
    assert tuple(other) == (0, 1, 2)


def test_dynamicenum():
    """Dynamic enum."""

    class MyType(u.DynamicEnumType):
        pass

    one = MyType()
    one.add(0, "a")
    one.add(1, "b")
    assert tuple(one) == (0, 1)

    other = MyType()
    other.add(2, "c")

    assert one is not other
    assert one != other
    assert tuple(one) == (0, 1)
    assert tuple(other) == (2,)


def test_cast():
    """Enum."""

    class OneType(u.AEnumType):
        """Enum."""

        keytype: u.AScalarType = u.UintType(4)

        def _build(self) -> None:
            self._add(u.AUTO, "a", title="title")

    class TwoType(u.AEnumType):
        """Enum."""

        keytype: u.AScalarType = u.UintType(4)

        def _build(self) -> None:
            self._add(u.AUTO, "a", title="title")

    class ThreeType(u.AEnumType):
        """Enum."""

        keytype: u.AScalarType = u.UintType(3)

        def _build(self) -> None:
            self._add(u.AUTO, "a", title="title")

    assert OneType().cast(TwoType()) == [
        (
            "",
            "",
        ),
    ]
    assert TwoType().cast(OneType()) == OneType().cast(TwoType())

    assert OneType().cast(ThreeType()) is None
    assert ThreeType().cast(OneType()) is None



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Configuration."""


def test_example_simple(example_simple):
    """Basic Testing."""
    from uart_lib.uart import UartMod

    mod = UartMod()

    assert mod.basename == "uart"
    assert mod.modname == "uart"
    assert mod.libname == "uart_lib"
    assert mod.qualname == "uart_lib.uart"
    assert [repr(item) for item in mod.namespace] == [
        "Port(ClkRstAnType(), 'main_i', direction=IN, doc=Doc(title='Clock and Reset'))",
        "Port(UartIoType(), 'uart_i', direction=IN, doc=Doc(title='UART', comment='RX/TX'), clkrel=ASYNC)",
        "Port(BusType(), 'bus_i', direction=IN, clkrel=ClkRel(clk=Port(ClkType(), "
        "'main_clk_i', direction=IN, doc=Doc(title='Clock'))))",
        "Signal(ClkType(), 'clk_s', doc=Doc(title='Clock'))",
        "Signal(DynamicStructType(), 'core_regf_i_s', direction=OUT)",
    ]
    assert [repr(item) for item in mod.params] == []
    assert [repr(item) for item in mod.ports] == [
        "Port(ClkRstAnType(), 'main_i', direction=IN, doc=Doc(title='Clock and Reset'))",
        "Port(UartIoType(), 'uart_i', direction=IN, doc=Doc(title='UART', comment='RX/TX'), clkrel=ASYNC)",
        "Port(BusType(), 'bus_i', direction=IN, clkrel=ClkRel(clk=Port(ClkType(), "
        "'main_clk_i', direction=IN, doc=Doc(title='Clock'))))",
    ]
    assert [repr(item) for item in mod.ports.leveliter()] == [
        "(0, Port(ClkRstAnType(), 'main_i', direction=IN, doc=Doc(title='Clock and Reset')))",
        "(1, Port(ClkType(), 'main_clk_i', direction=IN, doc=Doc(title='Clock')))",
        "(1, Port(RstAnType(), 'main_rst_an_i', direction=IN, doc=Doc(title='Async "
        "Reset', descr='Low-Active', comment='Async Reset (Low-Active)')))",
        "(0, Port(UartIoType(), 'uart_i', direction=IN, doc=Doc(title='UART', comment='RX/TX'), clkrel=ASYNC))",
        "(1, Port(BitType(), 'uart_rx_o', direction=OUT, clkrel=ASYNC))",
        "(1, Port(BitType(), 'uart_tx_i', direction=IN, clkrel=ASYNC))",
        "(0, Port(BusType(), 'bus_i', direction=IN, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(TransType(), 'bus_trans_i', direction=IN, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(AddrType(32), 'bus_addr_i', direction=IN, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(WriteType(), 'bus_write_i', direction=IN, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(DataType(32), 'bus_wdata_i', direction=IN, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(BitType(default=1), 'bus_ready_o', direction=OUT, "
        "doc=Doc(descr='Transfer is finished on HIGH.'), "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(RespType(), 'bus_resp_o', direction=OUT, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
        "(1, Port(DataType(32), 'bus_rdata_o', direction=OUT, "
        "clkrel=ClkRel(clk=Port(ClkType(), 'main_clk_i', direction=IN, "
        "doc=Doc(title='Clock')))))",
    ]
    assert [repr(item) for item in mod.portssignals] == [
        "Port(ClkRstAnType(), 'main_i', direction=IN, doc=Doc(title='Clock and Reset'))",
        "Port(UartIoType(), 'uart_i', direction=IN, doc=Doc(title='UART', comment='RX/TX'), clkrel=ASYNC)",
        "Port(BusType(), 'bus_i', direction=IN, clkrel=ClkRel(clk=Port(ClkType(), "
        "'main_clk_i', direction=IN, doc=Doc(title='Clock'))))",
        "Signal(ClkType(), 'clk_s', doc=Doc(title='Clock'))",
        "Signal(DynamicStructType(), 'core_regf_i_s', direction=OUT)",
    ]
    assert [repr(item) for item in mod.insts] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
    ]


def test_example_param(example_param):
    """Param Example."""
    from param_lib.param import ParamMod

    mod = ParamMod()
    assert [repr(item) for item in mod.namespace] == [
        "Param(IntegerType(default=10), 'param_p')",
        "Param(IntegerType(default=Log2Expr(Op(Param(IntegerType(default=10), "
        "'param_p'), '+', ConstExpr(IntegerType(default=1))))), 'width_p')",
        "Param(IntegerType(default=Param(IntegerType(default=10), 'param_p')), 'default_p')",
        "Port(UintType(Param(IntegerType(default=10), 'param_p')), 'data_i', direction=IN)",
        "Port(UintType(Param(IntegerType(default=Log2Expr(Op(Param(IntegerType(default=10), "
        "'param_p'), '+', ConstExpr(IntegerType(default=1))))), 'width_p')), "
        "'cnt_o', direction=OUT)",
        "Const(UintType(Param(IntegerType(default=10), 'param_p'), "
        "default=Param(IntegerType(default=Param(IntegerType(default=10), "
        "'param_p')), 'default_p')), 'const_c')",
    ]



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import re

from pytest import raises

import ucdp as u


class DutMod(u.ATailoredMod):
    """Example DUT."""

    def _build(self):
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(u.UintType(8), "data_i")
        self.add_port(u.UintType(8), "data_o")


class SubMod(u.AMod):
    """A Sub Module."""

    def _build(self):
        DutMod(self, "u_dut")


class TopMod(u.AMod):
    """A Top Module."""

    def _build(self):
        SubMod(self, "u_sub0")
        SubMod(self, "u_sub1")
        DutMod(self, "u_dut")


class TbMod(u.ATbMod):
    """A Testbench Module."""

    def _build(self):
        mod = DutMod(self, "u_dut")
        mod.con("main_i", "create(main_s)")


class GenTbMod(u.AGenericTbMod):
    """A Testbench Module."""

    dut_modclss: u.ClassVar[u.ModClss] = {DutMod, TopMod}

    @classmethod
    def build_dut(cls, **kwargs) -> u.BaseMod:
        """Build DUT."""
        return DutMod()  # type: ignore[call-arg]

    def _build(self):
        dut = self.dut
        dut.con("main_i", "create(main_s)")


class TopTbMod(u.ATbMod):
    """A Top Testbench Module."""

    dut_modclss: u.ClassVar[u.ModClss] = {TopMod}

    @classmethod
    def build_dut(cls, **kwargs) -> u.BaseMod:
        """Build DUT."""
        return TopMod()  # type: ignore[call-arg]

    def _build(self):
        pass


def test_basic():
    """Test Basics."""
    tb = TbMod()
    assert tb.modname == "tb"
    assert tb.topmodname == "tb"
    assert tb.libname == "tests"
    assert tb.is_tb is True
    assert repr(tb) == "<tests.test_modtb.TbMod(inst='tb', libname='tests', modname='tb')>"
    assert tb.get_modref() == u.ModRef("tests", "test_modtb", modclsname="TbMod")
    assert repr(tb.get_inst("u_dut")) == "<tests.test_modtb.DutMod(inst='tb/u_dut', libname='tests', modname='tb_dut')>"


def test_generic():
    """Test Basics."""
    tb = GenTbMod()
    dut = tb.dut
    assert isinstance(dut, DutMod)

    assert tb.modname == "gen_tb_dut"
    assert tb.topmodname == "gen_tb"
    assert tb.libname == "tests"
    assert tb.is_tb is True
    assert tuple(tb.get_tests()) == ()
    assert repr(tb) == (
        "<tests.test_modtb.GenTbMod(inst='gen_tb_dut', libname='tests', modname='gen_tb_dut', "
        "dut=<tests.test_modtb.DutMod(inst='dut', libname='tests', modname='dut')>)>"
    )
    assert tb.get_modref() == u.ModRef("tests", "test_modtb", modclsname="GenTbMod")

    assert dut.modname == "dut"
    assert dut.topmodname == "dut"
    assert dut.libname == "tests"
    assert dut.is_tb is False
    assert repr(dut) == "<tests.test_modtb.DutMod(inst='dut', libname='tests', modname='dut')>"
    assert dut.get_modref() == u.ModRef("tests", "test_modtb", modclsname="DutMod")


def test_wrong_mod():
    """Test Wrong Module."""
    msg = "<class 'tests.test_modtb.GenTbMod'> can only test"
    with raises(TypeError, match=re.escape(msg)):
        GenTbMod.build_tb(SubMod())



"""My Name Module."""


from fileliststandard import HdlFileList
from glbl_lib.bus import BusType
from glbl_lib.clk_gate import ClkGateMod
from glbl_lib.regf import RegfMod

import ucdp as u


class MyNameIoType(u.AStructType):
    """My Name IO."""

    title: str = "My Name"
    comment: str = "RX/TX"

    def _build(self) -> None:
        self._add("rx", u.BitType(), u.BWD)
        self._add("tx", u.BitType(), u.FWD)


class MyNameMod(u.AMod):
    """My Name Module."""

    filelists: u.ClassVar[u.ModFileLists] = (
        HdlFileList(gen="full"),
    )

    def _build(self) -> None:
        """Build."""
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(MyNameIoType(), "my_name_i", route="create(u_core/my_name_i)", clkrel=u.ASYNC)
        self.add_port(BusType(), "bus_i", clkrel="main_clk_i")

        clkgate = ClkGateMod(self, "u_clk_gate")
        clkgate.con("clk_i", "main_clk_i")
        clkgate.con("clk_o", "create(clk_s)")

        regf = RegfMod(self, "u_regf")
        regf.con("main_i", "main_i")
        regf.con("bus_i", "bus_i")

        core = MyNameCoreMod(parent=self, name="u_core")

        core.add_port(u.ClkRstAnType(), "main_i")
        core.con("main_clk_i", "clk_s")
        core.con("main_rst_an_i", "main_rst_an_i")
        core.con("create(regf_i)", "u_regf/regf_o")

        word = regf.add_word("ctrl")
        word.add_field("ena", u.EnaType(), is_readable=True, route="u_clk_gate/ena_i")
        word.add_field("strt", u.BitType(), is_writable=True, route="create(u_core/strt_i)")


class MyNameCoreMod(u.ACoreMod):
    """A Simple My Name."""

    filelists: u.ClassVar[u.ModFileLists] = (HdlFileList(gen="inplace"),)




#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import re

from pytest import raises

import ucdp as u


class MyConfig(u.AConfig):
    """A Configuration."""

    feature: bool = False
    size: int = 0


class MyMod(u.AConfigurableMod):
    """Example Configurable Module."""

    config: MyConfig

    def _build(self):
        config = self.config
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(u.UintType(8), "data_i")
        self.add_port(u.UintType(8), "data_o")
        if config.feature:
            self.add_port(u.UintType(8), "feat_i")
            self.add_port(u.UintType(8), "feat_o")


class MyOtherMod(MyMod):
    """Configurable Module with Other Config."""

    config: MyConfig = MyConfig(feature=True)


class AnotherMod(MyMod):
    """Configurable Module with Other Config."""

    config: MyConfig = MyConfig("foo")


def test_noconfig():
    """Configurable Module."""
    with raises(ValueError):
        MyMod()


def test_config():
    """Configurable Module."""
    mod = MyMod(config=MyConfig())
    assert mod.modname == "my"
    assert mod.topmodname == "my"

    config = mod.config
    assert config == MyConfig("")
    assert config.feature is False
    assert config.size == 0


class TopMod(u.AMod):
    """Instance without Config."""

    def _build(self) -> None:
        MyOtherMod(self, "u_inst")


def test_no_inst_config():
    """Config is required on Instance Creation."""
    msg = "'config' is required if 'parent' is given"
    with raises(ValueError, match=re.escape(msg)):
        TopMod()


def test_no_default_config():
    """Configurable Module."""
    mod = MyMod(config=MyConfig("name"))
    assert mod.modname == "my_name"
    assert mod.topmodname == "my"

    config = mod.config
    assert config == MyConfig("name")
    assert config.feature is False
    assert config.size == 0


def test_other():
    """Other Configurable Module."""
    mod = MyOtherMod()
    assert mod.modname == "my_other"
    assert mod.topmodname == "my"

    config = mod.config
    assert config == MyConfig(feature=True)
    assert config.feature is True
    assert config.size == 0


def test_another():
    """Another Configurable Module."""
    mod = AnotherMod()
    assert mod.modname == "another_foo"
    assert mod.topmodname == "my"


class SubConfig(u.AConfig):
    """Sub Configuration."""

    baseaddr: u.Hex = 0x1000


class SubMod(u.AConfigurableMod):
    """Sub Module."""

    def _build(self):
        pass


class RootMod(u.AMod):
    """Parent Module."""

    def _build(self):
        SubMod(self, "u_inst0", config=SubConfig())
        SubMod(self, "u_inst1", config=SubConfig(baseaddr=0x1000))
        SubMod(self, "u_inst2", config=SubConfig(baseaddr=0x2000))
        SubMod(self, "u_inst3", config=SubConfig(name="one", baseaddr=0x2000))


def test_sub():
    """Sub."""
    mod = RootMod()
    inst0 = mod.get_inst("u_inst0")
    assert inst0.modname == "sub"
    assert inst0.topmodname == "sub"

    inst1 = mod.get_inst("u_inst1")
    assert inst1.modname == "sub"
    assert inst1.topmodname == "sub"

    inst2 = mod.get_inst("u_inst2")
    assert inst2.modname == "sub_8a5068b003412f45"
    assert inst2.topmodname == "sub"

    inst3 = mod.get_inst("u_inst3")
    assert inst3.modname == "sub_one"
    assert inst3.topmodname == "sub"


class UniqueConfig(u.AConfig):
    """Configuration."""

    baseaddr: u.Hex = 0x1000

    @property
    def unique_name(self):
        """Unique Name."""
        return f"b{self.baseaddr!s}"


class UniqueMod(u.AConfigurableMod):
    """Example Configurable Module."""

    config: UniqueConfig = UniqueConfig()

    def _build(self):
        pass


def test_unique():
    """Unique Name."""
    mod = UniqueMod()
    assert mod.modname == "unique"

    mod = UniqueMod(config=UniqueConfig())
    assert mod.modname == "unique"

    mod = UniqueMod(config=UniqueConfig(baseaddr=0x2000))
    assert mod.modname == "unique_b0x2000"


def test_missing_config():
    """Missing Configuration."""
    with raises(u.ValidationError):
        MyMod()


class DefaultMod(u.AConfigurableMod):
    """Example Configurable Module."""

    config: UniqueConfig

    def _build(self):
        pass

    def get_default_config(self) -> UniqueConfig:
        """Calculate Config."""
        return UniqueConfig(baseaddr=0xCAFE)


def test_default_config():
    """Default Configuration."""
    mod = DefaultMod()
    assert mod.modname == "default"


def test_config_modname():
    """Modname from config."""
    config = MyConfig(modname="forced_name")
    mod = MyMod(config=config)
    assert mod.modname == "forced_name"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Type Specific Casts."""

import ucdp as u


def test_bit():
    """Bit Casting."""
    type_ = u.BitType()

    assert type_.cast(u.UintType(1)) == [("", "")]
    assert type_.cast(u.UintType(1, default=1)) == [("", "")]
    assert type_.cast(u.UintType(2)) is None

    assert type_.cast(u.SintType(1)) == [("", "")]
    assert type_.cast(u.SintType(1, default=-1)) == [("", "")]
    assert type_.cast(u.SintType(2)) is None

    assert type_.cast(u.IntegerType()) is None

    class MyEnum1(u.AEnumType):
        keytype: u.UintType = u.UintType(1)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum1()) == [("", "")]

    class MyEnum2(u.AEnumType):
        keytype: u.UintType = u.UintType(2)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum2()) is None


def test_rail():
    """Rail Casting."""
    type_ = u.RailType()

    assert type_.cast(u.UintType(1)) == [("", "")]
    assert type_.cast(u.UintType(1, default=1)) == [("", "")]
    assert type_.cast(u.UintType(2)) is None

    assert type_.cast(u.SintType(1)) == [("", "")]
    assert type_.cast(u.SintType(1, default=-1)) == [("", "")]
    assert type_.cast(u.SintType(2)) is None

    assert type_.cast(u.IntegerType()) is None

    class MyEnum1(u.AEnumType):
        keytype: u.UintType = u.UintType(1)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum1()) == [("", "")]

    class MyEnum2(u.AEnumType):
        keytype: u.UintType = u.UintType(2)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum2()) is None


def test_uint():
    """Uint Casting."""
    type_ = u.UintType(3)

    assert type_.cast(u.BitType()) is None

    assert type_.cast(u.SintType(3)) == [("", "")]
    assert type_.cast(u.SintType(3, default=-1)) == [("", "")]
    assert type_.cast(u.SintType(2)) is None

    assert type_.cast(u.IntegerType()) is None
    assert u.UintType(32).cast(u.IntegerType()) == [("", "")]

    class MyEnum3(u.AEnumType):
        keytype: u.UintType = u.UintType(3)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum3()) == [("", "")]

    class MyEnum2(u.AEnumType):
        keytype: u.UintType = u.UintType(2)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum2()) is None


def test_sint():
    """Sint Casting."""
    type_ = u.SintType(3)

    assert type_.cast(u.BitType()) is None

    assert type_.cast(u.UintType(3)) == [("", "")]
    assert type_.cast(u.UintType(3, default=1)) == [("", "")]
    assert type_.cast(u.UintType(2)) is None

    assert type_.cast(u.IntegerType()) is None
    assert u.SintType(32).cast(u.IntegerType()) == [("", "")]

    class MyEnum3(u.AEnumType):
        keytype: u.UintType = u.UintType(3)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum3()) == [("", "")]

    class MyEnum2(u.AEnumType):
        keytype: u.UintType = u.UintType(2)

        def _build(self) -> None:
            pass

    assert type_.cast(MyEnum2()) is None



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Configuration."""

import datetime
import hashlib
from typing import ClassVar

from .consts import PAT_OPT_IDENTIFIER
from .object import Field, LightObject, PosArgs


class AConfig(LightObject):
    """
    Configuration Container.

    Args:
        name: Configuration name, used as suffix of the generated module.

    A configuration is nothing more than a recipe how to assemble a module:

    * if a specific option should be built-in or not
    * how many instances or which instances should be created

    A configuration **MUST** have at least a name.

    Due to the frozen instance approach, configurations have to be implemented
    via `u.field()`.

    ??? Example "AConfig Examples"
        Create a Config.

            >>> import ucdp as u
            >>> class MyConfig(u.AConfig):
            ...
            ...     base_addr: u.Hex # required without default
            ...     ram_size: u.Bytesize
            ...     rom_size: u.Bytesize|None = None
            ...     feature: bool = False

        To create 1st variant

            >>> variant0  = MyConfig(name='variant0', base_addr=4*1024, ram_size='16kB')
            >>> variant0
            MyConfig('variant0', base_addr=Hex('0x1000'), ram_size=Bytesize('16 KB'))
            >>> variant0.base_addr
            Hex('0x1000')
            >>> variant0.ram_size
            Bytesize('16 KB')
            >>> variant0.rom_size
            >>> variant0.feature
            False
            >>> variant0.hash
            'd0cf21ab30478424'

        To create 2nd variant

            >>> for name, value in variant0:
            ...     name, value
            ('name', 'variant0')
            ('modname', '')
            ('base_addr', Hex('0x1000'))
            ('ram_size', Bytesize('16 KB'))
            ('rom_size', None)
            ('feature', False)

            >>> variant1  = MyConfig('variant1', base_addr=8*1024, rom_size="2KB", ram_size="4KB", feature=True)
            >>> variant1
            MyConfig('variant1', base_addr=Hex('0x2000'), ram_size=Bytesize('4 KB'), rom_size=Bytesize('2 KB'), ...)
            >>> variant1.base_addr
            Hex('0x2000')
            >>> variant1.ram_size
            Bytesize('4 KB')
            >>> variant1.rom_size
            Bytesize('2 KB')
            >>> variant1.feature
            True
            >>> variant1.hash
            'eb8052ba0b38cc3f'

        To create another variant based on an existing:

            >>> variant2 = variant1.new(name='variant2', rom_size='8KB')
            >>> variant2
            MyConfig('variant2', base_addr=Hex('0x2000'), ram_size=Bytesize('4 KB'), rom_size=Bytesize('8 KB'), ...)
            >>> variant2.base_addr
            Hex('0x2000')
            >>> variant2.ram_size
            Bytesize('4 KB')
            >>> variant2.rom_size
            Bytesize('8 KB')
            >>> variant2.feature
            True
            >>> variant2.hash
            'd8704dd8163a5ecc'

    ???+ bug "Todo"
        * fix name type
    """

    name: str = Field(pattern=PAT_OPT_IDENTIFIER, default="")
    """Configuration Name."""

    modname: str = Field(pattern=PAT_OPT_IDENTIFIER, default="")
    """Enforced Module Name."""

    _posargs: ClassVar[PosArgs] = ("name",)
    _hash_excludes: ClassVar[set[str]] = set()

    def __init__(self, name: str = "", **kwargs):
        super().__init__(name=name, **kwargs)  # type: ignore[call-arg]

    @property
    def is_default(self) -> bool:
        """Return `true` if configuration just contains default values."""
        exclude = self.__class__._hash_excludes
        return not self.model_dump(exclude_unset=True, exclude_defaults=True, exclude=exclude)

    @property
    def hash(self) -> str:
        """Unique Configuration Hash."""
        exclude = self.__class__._hash_excludes
        hashdata = self.model_dump(exclude=exclude)
        for key, value in self.__dict__.items():
            if key not in hashdata:
                continue
            if isinstance(value, AConfig):
                hashdata[key] = value.hash
            if isinstance(value, tuple):
                items = list(value)
                for idx, item in enumerate(items):
                    if isinstance(item, AConfig):
                        items[idx] = item.hash
                hashdata[key] = tuple(items)
        return hashlib.sha256(str(hashdata).encode("utf-8")).hexdigest()[:16]

    @property
    def unique_name(self) -> str:
        """Unique Configuration Hash."""
        return self.hash


BaseConfig = AConfig
"""BaseConfig"""


class AVersionConfig(AConfig):
    """
    Version Configuration Container.

    Attributes:
        title: Title.
        version: Version
        timestamp: Timestamp

    ??? Example "AVersionConfig Examples"
        Create a Config.

            >>> import ucdp as u
            >>> import datetime
            >>> class MyVersionConfig(u.AVersionConfig):
            ...     mem_baseaddr: u.Hex

            >>> version = MyVersionConfig(
            ...     'my',
            ...     title="Title",
            ...     version="1.2.3",
            ...     timestamp=datetime.datetime(2020, 10, 17, 23, 42),
            ...     mem_baseaddr=0x12340000
            ... )
            >>> version.name
            'my'
            >>> version.title
            'Title'
            >>> version.timestamp
            datetime.datetime(2020, 10, 17, 23, 42)
            >>> version.mem_baseaddr
            Hex('0x12340000')
            >>> version.hash
            'e98d9954882a5081'

            >>> for name, value in version:
            ...     name, value
            ('name', 'my')
            ('modname', '')
            ('title', 'Title')
            ('version', '1.2.3')
            ('timestamp', datetime.datetime(2020, 10, 17, 23, 42))
            ('mem_baseaddr', Hex('0x12340000'))

        Title, version and timestamp do not affect the hash

        >>> title = "Title"
        >>> version = "Version"
        >>> timestamp0 = datetime.datetime(2020, 10, 17, 23, 42)
        >>> timestamp1 = datetime.datetime(2020, 10, 17, 23, 43)
        >>> MyVersionConfig(mem_baseaddr=0x12340000, title=title, version=version, timestamp=timestamp0).hash
        '5b23b35be0ab5615'
        >>> MyVersionConfig(mem_baseaddr=0x12340000, title=title, version=version, timestamp=timestamp1).hash
        '5b23b35be0ab5615'

    """

    _hash_excludes: ClassVar[set[str]] = {"title", "version", "timestamp"}

    title: str
    version: str
    timestamp: datetime.datetime



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Ifdef Support."""

from typing import TypeAlias

from .consts import RE_IFDEF
from .define import Defines

Ifdefs: TypeAlias = tuple[str, ...]


def cast_ifdefs(value: Ifdefs | str | None) -> Ifdefs:
    """
    Cast Ifdefs.

    Examples:

        >>> cast_ifdefs('ASIC')
        ('ASIC',)
        >>> cast_ifdefs('!ASIC')
        ('!ASIC',)
        >>> cast_ifdefs(('ASIC',))
        ('ASIC',)
        >>> cast_ifdefs(('ASIC', 'BEHAV'))
        ('ASIC', 'BEHAV')
        >>> cast_ifdefs('')
        ()
        >>> cast_ifdefs(None)
        ()

    Forbidden:

        >>> cast_ifdefs('AS!')
        Traceback (most recent call last):
            ...
        ValueError: Invalid ifdef 'AS!'
        >>> cast_ifdefs('!!AS')
        Traceback (most recent call last):
            ...
        ValueError: Invalid ifdef '!!AS'
    """
    if not value:
        return ()
    if isinstance(value, str):
        value = (value,)
    if isinstance(value, tuple):
        for item in value:
            if not RE_IFDEF.match(item):
                raise ValueError(f"Invalid ifdef {item!r}")
        return value
    raise ValueError(f"Invalid ifdefs: {value}")


def join_ifdefs(base: Ifdefs, add: Ifdefs) -> Ifdefs:
    """
    Join Ifdefs.

    Examples:

        >>> join_ifdefs(('ASIC',), ('ASIC',))
        ('ASIC',)
        >>> join_ifdefs(('ASIC',), ('BEHAV',))
        ('ASIC', 'BEHAV')
        >>> join_ifdefs(('ASIC',), ('!ASIC',))
        ('!ASIC',)
        >>> join_ifdefs(('!ASIC',), ('ASIC',))
        ('ASIC',)
        >>> join_ifdefs(('ASIC', 'FOO'), ('!ASIC',))
        ('FOO', '!ASIC')
        >>> join_ifdefs(('ASIC', 'FOO'), ('!ASIC', 'FOO'))
        ('FOO', '!ASIC')
        >>> join_ifdefs(('ASIC', 'FOO'), ('!ASIC', 'BAR'))
        ('FOO', '!ASIC', 'BAR')
        >>> join_ifdefs(('!ASIC', 'FOO'), ('ASIC',))
        ('FOO', 'ASIC')
        >>> join_ifdefs(('!ASIC', 'FOO'), ('ASIC', 'FOO'))
        ('FOO', 'ASIC')
        >>> join_ifdefs(('!ASIC', 'FOO'), ('ASIC', 'BAR'))
        ('FOO', 'ASIC', 'BAR')
    """
    result = list(base)
    for ifdef in add:
        # remove inverse
        if ifdef.startswith("!"):
            if ifdef[1:] in result:
                result.remove(ifdef[1:])
        elif f"!{ifdef}" in result:
            result.remove(f"!{ifdef}")

        # add - if missing
        if ifdef not in result:
            result.append(ifdef)
    return tuple(result)


def resolve_ifdefs(defines: Defines, ifdefs: Ifdefs) -> Ifdefs | None:
    """
    Resolve Ifdefs.

    >>> import ucdp as u
    >>> defines = u.Defines([u.Define('FOO'), u.Define('BAR', value=3)])
    >>> resolve_ifdefs(defines, ())
    ()
    >>> resolve_ifdefs(defines, ('FOO',))
    ('FOO',)
    >>> resolve_ifdefs(defines, ('BAR',))
    ('BAR',)
    >>> resolve_ifdefs(defines, ('FOO', 'BAR'))
    ('FOO', 'BAR')
    >>> resolve_ifdefs(defines, ('FOO', '!BAR'))
    >>> resolve_ifdefs(defines, ('BAZ',))
    >>> resolve_ifdefs(defines, ('!BAZ',))
    ('!BAZ',)
    """
    result = []
    for ifdef in ifdefs:
        # abort if ifdef is not there OR
        # abort if ifdef is there but forbidden
        if (ifdef.removeprefix("!") in defines) == ifdef.startswith("!"):
            return None
        result.append(ifdef)
    return tuple(result)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Module Parameter .

??? Example "Parameter Examples"
    Usage:

        >>> from tabulate import tabulate
        >>> import ucdp as u
        >>> u.Param(u.UintType(6), "param_p")
        Param(UintType(6), 'param_p')

        Also complex types can simply be used:

        >>> class AType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("req", u.BitType())
        ...         self._add("data", u.ArrayType(u.UintType(16), 5))
        ...         self._add("ack", u.BitType(), u.BWD)
        >>> class MType(u.AEnumType):
        ...     keytype: u.AScalarType = u.UintType(2)
        ...     def _build(self) -> None:
        ...         self._add(0, "Linear")
        ...         self._add(1, "Cyclic")
        >>> class BType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("foo", AType())
        ...         self._add("mode", MType())
        ...         self._add("bar", u.ArrayType(AType(), 3), u.BWD)

    These types are automatically resolved by iterating over the parameter:

        >>> param = u.Param(BType(), "param_p")
        >>> for item in param:
        ...     print(repr(item))
        Param(BType(), 'param_p')
        Param(AType(), 'param_foo_p')
        Param(BitType(), 'param_foo_req_p')
        Param(ArrayType(UintType(16), 5), 'param_foo_data_p')
        Param(BitType(), 'param_foo_ack_p')
        Param(MType(), 'param_mode_p')
        Param(ArrayType(AType(), 3), 'param_bar_p')
        Param(ArrayType(BitType(), 3), 'param_bar_ack_p')
        Param(ArrayType(ArrayType(UintType(16), 5), 3), 'param_bar_data_p')
        Param(ArrayType(BitType(), 3), 'param_bar_req_p')

    Value:

        >>> for item in u.Param(u.UintType(6), "param_p").iter():
        ...     print(repr(item))
        Param(UintType(6), 'param_p')
        >>> for item in u.Param(u.UintType(6), "param_p").iter(value=42):
        ...     print(repr(item))
        Param(UintType(6, default=42), 'param_p')
"""

from typing import Any

from .ident import Ident, _iters


class Param(Ident):
    """
    Module Parameter.

    Args:
        type_ (AType): Type.
        name (Name): Name.

    Keyword Args:
        doc (Doc): Documentation Container
        value (Any): Value.


    Note:
        Parameter names should end with '_p'.

    ??? Example "Param Examples"
        Example:

            >>> import ucdp as u
            >>> cnt = u.Param(u.UintType(6), "cnt_p")
            >>> cnt
            Param(UintType(6), 'cnt_p')
            >>> cnt.type_
            UintType(6)
            >>> cnt.name
            'cnt_p'
            >>> cnt.basename
            'cnt'
            >>> cnt.suffix
            '_p'
            >>> cnt.doc
            Doc()
            >>> cnt.value

        If the parameter is casted via `int()` it returns `value` if set, other `type_.default`.

            >>> int(u.Param(u.UintType(6, default=2), "cnt_p"))
            2
            >>> int(u.Param(u.UintType(6, default=2), "cnt_p", value=4))
            4

        Parameter are Singleton:

            >>> u.Param(u.UintType(6), "cnt_p") is u.Param(u.UintType(6), "cnt_p")
            True
    """

    value: Any = None

    def __int__(self):
        value = self.value
        if value is None:
            value = self.type_.default
        return int(value or 0)

    def __iter__(self):
        for _, ident in _iters([self], value=self.value):
            yield ident



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Loader and Top."""

import re
from pathlib import Path

from caseconverter import pascalcase
from contextlib_chdir import chdir
from pytest import raises

import ucdp as u


def test_load_simple(example_simple):
    """Simple Module."""
    top = u.load("glbl_lib.clk_gate", paths=None)
    assert top.ref == u.TopModRef(u.ModRef("glbl_lib", "clk_gate"))
    assert top.mod.libname == "glbl_lib"
    assert top.mod.modname == "clk_gate"
    assert [str(mod) for mod in top.iter()] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='clk_gate', libname='glbl_lib', modname='clk_gate')>"
    ]
    assert [str(mod) for mod in top.get_mods()] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='clk_gate', libname='glbl_lib', modname='clk_gate')>"
    ]
    assert (
        str(top.get_mod("glbl_lib.clk_gate")) == "<glbl_lib.clk_gate.ClkGateMod(inst='clk_gate', "
        "libname='glbl_lib', modname='clk_gate')>"
    )

    retop = u.Top.from_mod(top.mod)
    assert retop.mod == top.mod
    assert retop.ref == u.TopModRef(u.ModRef("glbl_lib", "clk_gate", modclsname="ClkGateMod"))


def test_load_non_mod(example_bad):
    """Load Non Module."""
    with raises(ValueError) as exc:
        u.load("glbl_bad_lib.regf.MyMod", paths=None)
    assert str(exc.value) == "<class 'glbl_bad_lib.regf.MyMod'> is not a module aka child of <class ucdp.BaseMod>."


def test_load_non_top(example_simple):
    """Load Non-Top Module."""
    msg = "uart_lib.uart.UartCoreMod is not a top module as it bases on <class 'ucdp.modcore.ACoreMod'>"
    with raises(ValueError, match=re.escape(msg)):
        u.load("uart_lib.uart.UartCoreMod")


def test_load_complex(example_simple):
    """Complexer Module."""
    top = u.load("uart_lib.uart", paths=None)
    assert top.ref == u.TopModRef(u.ModRef("uart_lib", "uart"))
    assert top.mod.libname == "uart_lib"
    assert top.mod.modname == "uart"
    assert [repr(mod) for mod in top.iter()] == [
        "<uart_lib.uart.UartMod(inst='uart', libname='uart_lib', modname='uart')>",
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_regf/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
    ]
    assert [repr(mod) for mod in top.iter(unique=True)] == [
        "<uart_lib.uart.UartMod(inst='uart', libname='uart_lib', modname='uart')>",
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
    ]
    assert [repr(mod) for mod in top.iter(post=True)] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_regf/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
        "<uart_lib.uart.UartMod(inst='uart', libname='uart_lib', modname='uart')>",
    ]
    assert [repr(mod) for mod in top.iter(post=True, unique=True)] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
        "<uart_lib.uart.UartMod(inst='uart', libname='uart_lib', modname='uart')>",
    ]
    assert [repr(mod) for mod in top.get_mods()] == [repr(mod) for mod in top.iter(post=True)]
    assert [repr(mod) for mod in top.get_mods("uart_lib.uart*")] == [
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
        "<uart_lib.uart.UartCoreMod(inst='uart/u_core', libname='uart_lib', modname='uart_core')>",
        "<uart_lib.uart.UartMod(inst='uart', libname='uart_lib', modname='uart')>",
    ]
    assert (
        repr(top.get_mod("glbl_lib.clk_gate"))
        == "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>"
    )
    assert [repr(mod) for mod in top.get_mods("glbl_lib.clk_gate")] == [
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
        "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_regf/u_clk_gate', libname='glbl_lib', modname='clk_gate')>",
    ]
    assert [repr(mod) for mod in top.get_mods("glbl_lib.regf")] == []
    assert [repr(mod) for mod in top.get_mods("glbl_lib.regf", base=True)] == [
        "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>",
    ]

    msg = (
        "'glbl_lib.regf' not found. Known are:\n  glbl_lib.clk_gate\n  uart_lib.uart\n  "
        "uart_lib.uart_core\n  uart_lib.uart_regf"
    )
    with raises(ValueError, match=re.escape(msg)):
        top.get_mod("glbl_lib.regf")

    assert (
        repr(top.get_mod("glbl_lib.regf", base=True))
        == "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>"
    )

    msg = (
        "Found multiple hardware modules for 'uart_lib.*':\n  uart_lib.uart\n  uart_lib.uart_core\n  uart_lib.uart_regf"
    )
    with raises(ValueError, match=re.escape(msg)):
        top.get_mod("uart_lib.*")

    msg = "'glbl_lib.foo' not found. Known are:\n  glbl_lib.clk_gate\n  glbl_lib.regf\n  uart_lib.uart"
    with raises(ValueError, match=re.escape(msg)):
        top.get_mod("glbl_lib.foo", base=True)


def test_load_complex_sub(example_simple):
    """Complexer Module with Sub module."""
    top = u.load("uart_lib.uart-glbl_lib.clk_gate", paths=None)
    assert top.ref == u.TopModRef(u.ModRef("uart_lib", "uart"), sub="glbl_lib.clk_gate")
    assert (
        repr(top.mod)
        == "<glbl_lib.clk_gate.ClkGateMod(inst='uart/u_clk_gate', libname='glbl_lib', modname='clk_gate')>"
    )


def test_load_tb(example_simple):
    """Complexer Module with Testbench."""
    top = u.load("glbl_lib.regf_tb#uart_lib.uart-uart_lib.uart_regf", paths=None)
    assert top.ref == u.TopModRef(
        u.ModRef("uart_lib", "uart"), sub="uart_lib.uart_regf", tb=u.ModRef("glbl_lib", "regf_tb")
    )
    assert repr(top.mod) == (
        "<glbl_lib.regf_tb.RegfTbMod(inst='regf_tb_uart_regf', libname='glbl_lib', modname='regf_tb_uart_regf', "
        "dut=<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>)>"
    )
    assert repr(top.mod.dut) == "<glbl_lib.regf.RegfMod(inst='uart/u_regf', libname='uart_lib', modname='uart_regf')>"

    retop = u.Top.from_mod(top.mod)
    assert retop.mod == top.mod
    assert retop.ref == u.TopModRef(
        u.ModRef("glbl_lib", "regf", modclsname="RegfMod"), tb=u.ModRef("glbl_lib", "regf_tb", modclsname="RegfTbMod")
    )


def test_load_non_tb(example_simple):
    """Complexer Module with Testbench - Non-TB."""
    with raises(ValueError) as exc:
        u.load("glbl_lib.regf#uart_lib.uart-uart_lib.uart_regf", paths=None)
    assert (
        str(exc.value)
        == "<class 'glbl_lib.regf.RegfMod'> is not a testbench module aka child of <class ucdp.AGenericTbMod>."
    )


def test_imp_err(testdata):
    """Import Error."""
    msg = "'imp_err_lib.not_existing' not found."
    with raises(NameError, match=re.escape(msg)):
        u.load("imp_err_lib.not_existing", paths=None)


def test_imp_err_dep(testdata):
    """Broken Dependency."""
    msg = "No module named 'imp_err_lib.not_existing'"
    with raises(ModuleNotFoundError, match=re.escape(msg)):
        u.load("imp_err_lib.imp_err", paths=None)


def test_typo(example_simple):
    """Typos."""
    with raises(NameError, match=re.escape("'glbl_li.clk_gate' not found. Did you mean 'glbl_lib.clk_gate'")):
        u.load("glbl_li.clk_gate", paths=None)

    with raises(NameError, match=re.escape("'glbl_lib.clk_ate' not found. Did you mean 'glbl_lib.clk_gate'")):
        u.load("glbl_lib.clk_ate", paths=None)

    msg = "'glbl_lib.clk_gate' does not contain GateMod. Did you mean 'glbl_lib.clk_gate', '"
    with raises(NameError, match=re.escape(msg)):
        u.load("glbl_lib.clk_gate.GateMod", paths=None)


def test_paths(tmp_path):
    """Paths Argument."""
    for name in ("one", "two", "three", "four", "five"):
        one_path = tmp_path / name / f"{name}_lib"
        one_path.mkdir(parents=True)
        (one_path / f"{name}.py").write_text(f"""\
import ucdp as u

class {pascalcase(name)}Mod(u.AMod):
    def _build(self) -> None:
        pass
""")

    with raises(NameError):
        assert u.load("one_lib.one")

    # relative path
    with chdir(tmp_path):
        assert u.load("one_lib.one", paths=(Path("one"),)).mod.name == "one"

    with raises(NameError):
        assert u.load("two_lib.two")

    # absolute path
    with chdir(tmp_path):
        assert u.load("two_lib.two", paths=(tmp_path / "two",)).mod.name == "two"

    # glob
    with chdir(tmp_path):
        assert u.load("four_lib.four", paths=(tmp_path / "*",)).mod.name == "four"
